"""
Implements a LEGEND Data Object representing a variable-length array of
variable-length arrays and corresponding utilities.
"""

from __future__ import annotations

import logging
from collections.abc import Iterator, Mapping, Sequence
from typing import Any

import awkward as ak
import awkward_pandas as akpd
import numpy as np
import pandas as pd
from numba import jit
from numpy.typing import ArrayLike, DTypeLike, NDArray

from .. import utils
from . import arrayofequalsizedarrays as aoesa
from . import vovutils
from .array import Array
from .lgdo import LGDOCollection

log = logging.getLogger(__name__)


class VectorOfVectors(LGDOCollection):
    """A n-dimensional variable-length 1D array of variable-length 1D arrays.

    If the vector is 2-dimensional, the internal representation is as two NumPy
    arrays, one to store the flattened data contiguosly
    (:attr:`flattened_data`) and one to store the cumulative sum of lengths of
    each vector (:attr:`cumulative_length`). When the dimension is more than 2,
    :attr:`flattened_data` is a :class:`VectorOfVectors` itself.

    Examples
    --------
    >>> from lgdo import VectorOfVectors
    >>> data = VectorOfVectors(
    ...   [[[1, 2], [3, 4, 5]], [[2], [4, 8, 9, 7]], [[5, 3, 1]]],
    ...   attrs={"units": "m"}
    ... )
    >>> print(data)
    [[[1, 2], [3, 4, 5]],
     [[2], [4, 8, 9, 7]],
     [[5, 3, 1]]
    ] with attrs={'units': 'm'}
    >>> data.view_as("ak")
    <Array [[[1, 2], [3, 4, 5]], ..., [[5, ..., 1]]] type='3 * var * var * int64'>

    Note
    ----
    Many class methods are currently implemented only for 2D vectors and will
    raise an exception on higher dimensional data.
    """

    def __init__(
        self,
        data: ArrayLike | None = None,
        flattened_data: ArrayLike | None = None,
        cumulative_length: ArrayLike | VectorOfVectors | None = None,
        shape_guess: Sequence[int, ...] | None = None,
        dtype: DTypeLike | None = None,
        fill_val: int | float | None = None,
        attrs: Mapping[str, Any] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        data
            Any array-like structure accepted by the :class:`ak.Array`
            constructor, with the exception that elements cannot be of type
            ``OptionType``, ``UnionType`` or ``RecordType``.  Takes priority
            over `flattened_data` and `cumulative_length`. The serialization of
            the :class:`ak.Array` is performed through :func:`ak.to_buffers`.
            Since the latter returns non-data-owning NumPy arrays, which would
            prevent later modifications like resizing, a copy is performed.
        flattened_data
            if not ``None``, used as the internal array for
            `self.flattened_data`.  Otherwise, an internal `flattened_data` is
            allocated based on `cumulative_length` (or `shape_guess`) and `dtype`.
        cumulative_length
            if not ``None``, used as the internal array for
            `self.cumulative_length`. Should be `dtype` :any:`numpy.uint32`. If
            `cumulative_length` is ``None``, an internal `cumulative_length` is
            allocated based on the first element of `shape_guess`.
        shape_guess
            a NumPy-format shape specification, required if either of
            `flattened_data` or `cumulative_length` are not supplied.  The
            first element should not be a guess and sets the number of vectors
            to be stored. The second element is a guess or approximation of the
            typical length of a stored vector, used to set the initial length
            of `flattened_data` if it was not supplied.
        dtype
            sets the type of data stored in `flattened_data`. Required if
            `flattened_data` and `array` are ``None``.
        fill_val
            fill all of `self.flattened_data` with this value.
        attrs
            a set of user attributes to be carried along with this LGDO.
        """
        # sanitize
        if cumulative_length is not None and not isinstance(cumulative_length, Array):
            cumulative_length = Array(cumulative_length)
        if flattened_data is not None and not isinstance(
            flattened_data, (Array, VectorOfVectors)
        ):
            flattened_data = Array(flattened_data)

        if data is not None:
            if not isinstance(data, ak.Array):
                data = ak.Array(data)

            if data.ndim < 2:
                msg = (
                    "cannot initialize a VectorOfVectors with "
                    f"{data.ndim}-dimensional data"
                )
                raise ValueError(msg)

            # make sure it's not a record array
            if not vovutils._ak_is_valid(data):
                msg = "input array type is not supported!"
                raise ValueError(msg)

            # array might be non-jagged! ('container' will hold a ndim NumPy array)
            if not vovutils._ak_is_jagged(data):
                data = ak.from_regular(data, axis=None)

            # ak.to_buffer helps in de-serialization
            # NOTE: ak.to_packed() needed?
            form, _, container = ak.to_buffers(ak.to_packed(data))

            # check if bytestring
            curr = form
            for _ in range(data.ndim - 1):
                curr = curr.content
            if (
                "__array__" in curr.parameters
                and curr.parameters["__array__"] == "bytestring"
            ):
                diffs = np.diff(container[f"node{data.ndim - 1}-offsets"])
                if (diffs != diffs[0]).all():
                    err_msg = "Non uniform string lengths not supported"
                    raise NotImplementedError(err_msg)
                flattened_data = np.asarray(
                    ak.enforce_type(
                        ak.unflatten(
                            container.pop(
                                f"node{data.ndim}-data", np.empty(0, dtype=dtype)
                            ),
                            diffs[0],
                        ),
                        "bytes",
                    )
                )

                # if user-provided dtype is different than dtype from Awkward, cast
                # NOTE: makes a copy only if needed
                flattened_data = np.asarray(flattened_data, dtype=dtype)
            else:
                # NOTE: node#-data is not even in the dict if the awkward array is empty
                # NOTE: if the data arg was a numpy array, to_buffers() preserves
                # the original dtype
                # FIXME: have to copy the buffers, otherwise self will not own the
                # data and self.resize() will fail. Is it possible to avoid this?
                flattened_data = np.copy(
                    container.pop(f"node{data.ndim - 1}-data", np.empty(0, dtype=dtype))
                )

                # if user-provided dtype is different than dtype from Awkward, cast
                # NOTE: makes a copy only if needed
                flattened_data = np.asarray(flattened_data, dtype=dtype)

            # start from innermost VoV and build nested structure
            for i in range(data.ndim - 2, -1, -1):
                # NOTE: remember, omit the leading 0 from ak.Array offsets
                cumulative_length = np.copy(container[f"node{i}-offsets"][1:])

                if i != 0:
                    # at the beginning of the loop: initialize innermost
                    # flattened_data and replace current flattened_data
                    # reference. in the following iterations flattened_data is
                    # a VectorOfVectors
                    flattened_data = VectorOfVectors(
                        flattened_data=flattened_data,
                        cumulative_length=cumulative_length,
                    )

                else:
                    # at end we need to initialize self with the latest flattened_data
                    self.__init__(
                        flattened_data=flattened_data,
                        cumulative_length=cumulative_length,
                    )

        else:
            self.flattened_data = None
            self.cumulative_length = None

            # let's first setup cumulative_length...
            if cumulative_length is None:
                # initialize based on shape_guess
                if shape_guess is None:
                    # just make an empty 2D vector
                    shape_guess = (0, 0)

                # sanity check
                if len(shape_guess) < 2:
                    msg = "shape_guess must be a sequence of 2 integers or more"
                    raise ValueError(msg)

                # let's Awkward do the job here, we're lazy
                if fill_val is not None:
                    self.__init__(
                        np.full(shape=shape_guess, fill_value=fill_val, dtype=dtype)
                    )
                else:
                    self.__init__(np.empty(shape=shape_guess, dtype=dtype))
            else:
                # if it's user provided just use it
                self.cumulative_length = cumulative_length

            # ...then flattened_data
            # NOTE: self.flattened_data might have already been initialized
            # above
            if flattened_data is None and self.flattened_data is None:
                # this happens when the cumulative_length arg is not None
                if dtype is None:
                    msg = "flattened_data and dtype cannot both be None!"
                    raise ValueError(msg)

                # now ready to initialize the object!
                self.flattened_data = Array(
                    shape=(self.cumulative_length[-1],), dtype=dtype, fill_val=fill_val
                )
            elif self.flattened_data is None:
                self.flattened_data = flattened_data

        super().__init__(attrs)

    @property
    def ndim(self):
        return 1 + (
            1 if isinstance(self.flattened_data, Array) else self.flattened_data.ndim
        )

    @property
    def dtype(self) -> np.dtype:
        return self.flattened_data.dtype

    def datatype_name(self) -> str:
        return "array"

    def form_datatype(self) -> str:
        eltype = (
            "array<1>{" + utils.get_element_type(self) + "}"
            if self.ndim == 2
            else self.flattened_data.form_datatype()
        )
        return "array<1>{" + eltype + "}"

    def __len__(self) -> int:
        """Return the number of stored vectors along the first axis (0)."""
        return len(self.cumulative_length)

    def __eq__(self, other: VectorOfVectors) -> bool:
        if isinstance(other, VectorOfVectors):
            if self.ndim == 2 and len(self.cumulative_length) != 0:
                fldata_eq = np.array_equal(
                    self.flattened_data[: self.cumulative_length[-1]],
                    other.flattened_data[: other.cumulative_length[-1]],
                )
            else:
                fldata_eq = self.flattened_data == other.flattened_data

            return (
                self.cumulative_length == other.cumulative_length
                and fldata_eq
                and self.dtype == other.dtype
                and self.attrs == other.attrs
            )

        return False

    def __getitem__(self, i: int) -> NDArray:
        """Return a view of the vector at index `i` along the first axis."""
        if self.ndim == 2:
            stop = self.cumulative_length[i]
            if i in (0, -len(self)):
                return self.flattened_data[0:stop]

            return self.flattened_data[self.cumulative_length[i - 1] : stop]

        raise NotImplementedError

    def __setitem__(self, i: int, new: NDArray) -> None:
        if self.ndim == 2:
            self.__getitem__(i)[:] = new
        else:
            raise NotImplementedError

    def reserve_capacity(self, cap_cl, *cap_args) -> None:
        """Set capacity of internal data arrays. Expect number of args to
        equal `self.n_dim`. First arg is capacity of cumulative length array.
        If `self.n_dim` is 2, second argument is capacity of flattened data,
        otherwise arguments are fed recursively to remaining dimensions.
        """
        self.cumulative_length.reserve_capacity(cap_cl)
        self.flattened_data.reserve_capacity(*cap_args)

    def get_capacity(self) -> tuple[int]:
        """Get tuple containing capacity of each dimension. First dimension
        is cumulative length array. Last dimension is flattened data.
        """
        fd_cap = self.flattened_data.get_capacity()
        if isinstance(fd_cap, int):
            return (self.cumulative_length.get_capacity(), fd_cap)
        return (self.cumulative_length.get_capacity(), *fd_cap)

    def trim_capacity(self) -> None:
        "Set capacity for all dimensions to minimum needed to hold data"
        self.cumulative_length.trim_capacity()
        self.flattened_data.trim_capacity()

    def resize(self, new_size: int, trim: bool = False) -> None:
        """Resize vector along the first axis.

        `self.flattened_data` is resized only if `new_size` is smaller than the
        current vector length.

        If `new_size` is larger than the current vector length,
        `self.cumulative_length` is padded with its last element.  This
        corresponds to appending empty vectors.

        If `trim` is ``True``, resize capacity to match new size

        Examples
        --------
        >>> vov = VectorOfVectors([[1, 2, 3], [4, 5]])
        >>> vov.resize(3)
        >>> print(vov)
        [[1 2 3],
         [4 5],
         [],
        ]

        >>> vov = VectorOfVectors([[1, 2], [3], [4, 5]])
        >>> vov.resize(2)
        >>> print(vov)
        [[1 2],
         [3],
        ]
        """
        old_s = len(self)

        # first resize the cumulative length
        self.cumulative_length.resize(new_size, trim)

        # if new_size > size, new elements are filled with zeros, let's fix
        # that
        if new_size > old_s:
            self.cumulative_length[old_s:] = self.cumulative_length[old_s - 1]

        # then resize the data array
        # if dlen > 0 this has no effect
        if len(self.cumulative_length) > 0:
            self.flattened_data.resize(self.cumulative_length[-1], trim)
        else:
            self.flattened_data.resize(0, trim)

    def append(self, new: NDArray) -> None:
        """Append a 1D vector `new` at the end.

        Examples
        --------
        >>> vov = VectorOfVectors([[1, 2, 3], [4, 5]])
        >>> vov.append([8, 9])
        >>> print(vov)
        [[1 2 3],
         [4 5],
         [8 9],
        ]
        """
        self.insert(len(self), new)

    def insert(self, i: int, new: NDArray) -> None:
        """Insert a vector at index `i`.

        `self.flattened_data` (and therefore `self.cumulative_length`) is
        resized in order to accommodate the new element.

        Examples
        --------
        >>> vov = VectorOfVectors([[1, 2, 3], [4, 5]])
        >>> vov.insert(1, [8, 9])
        >>> print(vov)
        [[1 2 3],
         [8 9],
         [4 5],
        ]
        """
        if self.ndim == 2:
            if i > len(self):
                msg = f"index {i} is out of bounds for vector with size {len(self)}"
                raise IndexError(msg)

            i_start = 0 if i == 0 else self.cumulative_length[i - 1]
            self.flattened_data.insert(i_start, new)
            self.cumulative_length.insert(i, i_start)
            self.cumulative_length[i:] += np.uint32(len(new))
        else:
            raise NotImplementedError

    def replace(self, i: int, new: NDArray) -> None:
        """Replace the vector at index `i` with `new`.

        `self.flattened_data` (and therefore `self.cumulative_length`) is
        resized, if the length of `new` is different from the vector currently
        at index `i`.

        Examples
        --------
        >>> vov = VectorOfVectors([[1, 2, 3], [4, 5]])
        >>> vov.replace(0, [8, 9])
        >>> print(vov)
        [[8 9],
         [4 5],
        ]
        """
        if self.ndim == 2:
            if i >= len(self):
                msg = f"index {i} is out of bounds for vector with size {len(self)}"
                raise IndexError(msg)

            vidx = self.cumulative_length
            dlen = len(new) - len(self[i])

            if dlen != 0:
                # move the subsequent entries
                vidx[i:] += dlen
                self.flattened_data.resize(vidx[-1])
                self.flattened_data._nda[vidx[i] : vidx[-1]] = self.flattened_data._nda[
                    vidx[i] - dlen : vidx[-1] - dlen
                ]

            # set the already allocated indices
            start = vidx[i - 1] if i > 0 else 0
            self.flattened_data[start : vidx[i]] = new
        else:
            raise NotImplementedError

    def _set_vector_unsafe(
        self, i: int, vec: NDArray, lens: ArrayLike | None = None
    ) -> None:
        r"""Insert vector `vec` at position `i`.

        Assumes that ``j = self.cumulative_length[i-1]`` is the index (in
        `self.flattened_data`) of the end of the `(i-1)`\ th vector and copies
        `vec` in ``self.flattened_data[j:sum(lens)]``. Finally updates
        ``self.cumulative_length[i]`` with the new flattened data array length.

        Vectors stored after index `i` can be overridden, producing unintended
        behavior. This method is typically used for fast sequential fill of a
        pre-allocated vector of vectors.

        If i`vec` is 1D array and `lens` is ``None``, set using full array. If
        `vec` is 2D, require `lens` to be included, and fill each array only up
        to lengths in `lens`.

        Danger
        ------
        This method can lead to undefined behavior or vector invalidation if
        used improperly. Use it only if you know what you are doing.

        See Also
        --------
        append, replace, insert
        """
        if self.ndim == 2:
            # check if current vector is empty and get the start index in
            # flattened_data
            start = 0 if i == 0 else self.cumulative_length[i - 1]

            # if the new element is 1D, convert to dummy 2D
            if len(vec.shape) == 1:
                vec = np.expand_dims(vec, axis=0)
                if lens is None:
                    lens = np.array([vec.shape[1]], dtype="u4")

            # this in case lens is 02, convert to 1D
            if not isinstance(lens, np.ndarray):
                lens = np.array([lens], dtype="u4")

            # calculate stop index in flattened_data
            cum_lens = np.add(start, lens.cumsum(), dtype=int)

            # fill with fast vectorized routine
            if np.issubdtype(self.flattened_data.dtype, np.unsignedinteger):
                nan_val = np.iinfo(self.flattened_data.dtype).max
            if np.issubdtype(self.flattened_data.dtype, np.integer):
                nan_val = np.iinfo(self.flattened_data.dtype).min
            else:
                nan_val = np.nan
            vovutils._nb_fill(
                vec,
                lens,
                np.array([nan_val]).astype(self.flattened_data.nda.dtype),
                self.flattened_data.nda[start : cum_lens[-1]],
            )

            # add new vector(s) length to cumulative_length
            self.cumulative_length[i : i + len(lens)] = cum_lens
        else:
            raise NotImplementedError

    def __iter__(self) -> Iterator[NDArray]:
        if self.ndim == 2:
            for j, stop in enumerate(self.cumulative_length):
                if j == 0:
                    yield self.flattened_data[0:stop]
                else:
                    yield self.flattened_data[self.cumulative_length[j - 1] : stop]
        else:
            raise NotImplementedError

    def __str__(self) -> str:
        string = self.view_as("ak").show(stream=None)

        string = str(string).strip().removesuffix("]")
        string += "\n]"

        tmp_attrs = self.attrs.copy()
        tmp_attrs.pop("datatype")
        if len(tmp_attrs) > 0:
            string += f" with attrs={tmp_attrs}"

        return string

    def __repr__(self) -> str:
        npopt = np.get_printoptions()
        np.set_printoptions(threshold=5, edgeitems=2, linewidth=100)
        out = (
            "VectorOfVectors(flattened_data="
            + repr(self.flattened_data)
            + ", cumulative_length="
            + repr(self.cumulative_length)
            + ", attrs="
            + repr(self.attrs)
            + ")"
        )
        np.set_printoptions(**npopt)
        return out

    def to_aoesa(
        self,
        max_len: int | None = None,
        fill_val: bool | int | float = np.nan,
        preserve_dtype: bool = False,
    ) -> aoesa.ArrayOfEqualSizedArrays:
        """Convert to :class:`ArrayOfEqualSizedArrays`.

        Note
        ----
        The dtype of the original vector is typically not strictly preserved.
        The output dtype will be either :class:`np.float64` or :class:`np.int64`.
        If you want to use the same exact dtype, set `preserve_dtype` to
        ``True``.

        Parameters
        ----------
        max_len
            the length of the returned array along its second dimension. Longer
            vectors will be truncated, shorter will be padded with `fill_val`.
            If ``None``, the length will be equal to the length of the longest
            vector.
        fill_val
            value used to pad shorter vectors up to `max_len`. The dtype of the
            output array will be such that both `fill_val` and the vector
            values can be represented in the same data structure.
        preserve_dtype
            whether the output array should have exactly the same dtype as the
            original vector of vectors. The type `fill_val` must be a
            compatible one.
        """
        if self.ndim == 2:
            if max_len is None:
                lens = np.copy(self.cumulative_length)
                lens[1:] = lens[1:] - lens[:-1]
                max_len = int(np.max(lens))
            nda = np.full((len(self), max_len), fill_val)
            if preserve_dtype:
                nda = nda.astype(self.flattened_data.dtype, copy=False)

            _to_aoesa(self.flattened_data.nda, self.cumulative_length.nda, nda)

            return aoesa.ArrayOfEqualSizedArrays(nda=nda, attrs=self.getattrs())

        raise NotImplementedError

    def view_as(
        self,
        library: str,
        with_units: bool = False,
        fill_val: bool | int | float = np.nan,
        preserve_dtype: bool = False,
    ) -> pd.DataFrame | np.NDArray | ak.Array:
        r"""View the vector data as a third-party format data structure.

        This is typically a zero-copy or nearly zero-copy operation.

        Supported third-party formats are:

        - ``pd``: returns a :class:`pandas.Series` (supported through the
          ``awkward-pandas`` package)
        - ``np``: returns a :class:`numpy.ndarray`, padded with zeros to make
          it rectangular. This implies memory re-allocation.
        - ``ak``: returns an :class:`ak.Array`. ``self.cumulative_length`` is
          currently re-allocated for technical reasons.

        Notes
        -----
        Awkward array views partially involve memory re-allocation (the
        `cumulative_length`\ s), while NumPy "exploded" views clearly imply a
        full copy.

        Parameters
        ----------
        library
            format of the returned data view.
        with_units
            forward physical units to the output data.
        fill_val
            forwarded to :meth:`.to_aoesa`, if `library` is ``np``.
        preserve_dtype
            forwarded to :meth:`.to_aoesa`, if `library` is ``np``.

        See Also
        --------
        .LGDO.view_as
        """
        attach_units = with_units and "units" in self.attrs

        if library == "ak":
            if attach_units:
                msg = "Pint does not support Awkward yet, you must view the data with_units=False"
                raise ValueError(msg)

            # see https://github.com/scikit-hep/awkward/discussions/2848

            # cannot avoid making a copy here. we should add the leading 0 to
            # cumulative_length inside VectorOfVectors at some point in the
            # future
            offsets = np.empty(
                len(self.cumulative_length) + 1, dtype=self.cumulative_length.dtype
            )
            offsets[1:] = self.cumulative_length.nda
            offsets[0] = 0

            if self.ndim != 2:
                content = self.flattened_data.view_as(
                    library, with_units=with_units
                ).layout
            # need to handle strings separately
            elif np.issubdtype(self.flattened_data.nda.dtype, np.bytes_):
                byte_arrays = []
                for s in self.flattened_data.nda:
                    # Convert each string to array of bytes
                    byte_array = np.frombuffer(s, dtype=np.uint8)
                    byte_arrays.append(byte_array)
                max_len = max(len(b) for b in byte_arrays)
                raw_arrays = ak.contents.NumpyArray(np.concatenate(byte_arrays))
                array_of_chars = ak.contents.RegularArray(
                    raw_arrays, max_len, parameters={"__array__": "bytes"}
                )
                content = ak.enforce_type(array_of_chars, "bytes", highlevel=False)
            else:
                content = ak.contents.NumpyArray(self.flattened_data.nda)

            layout = ak.contents.ListOffsetArray(
                offsets=ak.index.Index(offsets),
                content=content,
            )
            return ak.Array(layout)

        if library == "np":
            if preserve_dtype:
                return self.to_aoesa(fill_val=fill_val, preserve_dtype=True).view_as(
                    "np", with_units=with_units
                )

            return self.to_aoesa().view_as("np", with_units=with_units)

        if library == "pd":
            if attach_units:
                msg = "Pint does not support Awkward yet, you must view the data with_units=False"
                raise ValueError(msg)

            return akpd.from_awkward(self.view_as("ak"))

        msg = f"{library} is not a supported third-party format."
        raise ValueError(msg)


@jit
def _to_aoesa(flattened_array, cumulative_length, nda):
    prev_cl = 0
    for i, cl in enumerate(cumulative_length):
        nda[i, : (cl - prev_cl)] = flattened_array[prev_cl:cl]
        prev_cl = cl



"""
Implements a LEGEND Data Object representing a struct and corresponding
utilities.
"""

from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Any

import numpy as np

from .lgdo import LGDO

log = logging.getLogger(__name__)


class Struct(LGDO, dict):
    """A dictionary of LGDO's with an optional set of attributes.

    After instantiation, add fields using :meth:`add_field` to keep the
    datatype updated, or call :meth:`update_datatype` after adding.
    """

    def __init__(
        self,
        obj_dict: Mapping[str, LGDO] | None = None,
        attrs: Mapping[str, Any] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        obj_dict
            instantiate this Struct using the supplied named LGDO's.  Note: no
            copy is performed, the objects are used directly.
        attrs
            a set of user attributes to be carried along with this LGDO.
        """
        if obj_dict is not None:
            for k, v in obj_dict.items():
                # check if value is another mapping-like object
                # initialize another struct (or derived class) in such a case
                if not isinstance(v, LGDO) and isinstance(v, Mapping):
                    # NOTE: calling self.__new__() and then self.__init__() allows for polymorphism
                    # but is there a better way?
                    nested = self.__new__(type(self), v)
                    nested.__init__(v)
                    super().update({k: nested})
                else:
                    # otherwise object must be an LGDO
                    if not isinstance(v, LGDO):
                        msg = f"value of '{k}' ({v!r}) is not an LGDO or a dictionary"
                        raise ValueError(msg)

                    # assign
                    super().update({k: v})

        # call LGDO constructor to setup attributes
        super().__init__(attrs)

    def datatype_name(self) -> str:
        return "struct"

    def form_datatype(self) -> str:
        return (
            self.datatype_name() + "{" + ",".join([str(k) for k in self.keys()]) + "}"
        )

    def update_datatype(self) -> None:
        self.attrs["datatype"] = self.form_datatype()

    def add_field(self, name: str | int, obj: LGDO) -> None:
        """Add a field to the table."""
        super().__setitem__(name, obj)
        self.update_datatype()

    def __setitem__(self, name: str, obj: LGDO) -> None:
        return self.add_field(name, obj)

    def __getattr__(self, name: str) -> LGDO:
        if hasattr(super(), name):
            return super().__getattr__(name)

        if name in self.keys():
            return super().__getitem__(name)

        raise AttributeError(name)

    def remove_field(self, name: str | int, delete: bool = False) -> None:
        """Remove a field from the table.

        Parameters
        ----------
        name
            name of the field to be removed.
        delete
            if ``True``, delete the field object by calling :any:`del`.
        """
        if delete:
            del self[name]
        else:
            self.pop(name)
        self.update_datatype()

    def __str__(self) -> str:
        """Convert to string (e.g. for printing)."""

        thr_orig = np.get_printoptions()["threshold"]
        np.set_printoptions(threshold=8)

        string = "{\n"
        for k, v in self.items():
            if "\n" in str(v):
                rv = str(v).replace("\n", "\n    ")
                string += f" '{k}':\n    {rv},\n"
            else:
                string += f" '{k}': {v},\n"
        string += "}"

        attrs = self.getattrs()
        if attrs:
            string += f" with attrs={attrs}"

        np.set_printoptions(threshold=thr_orig)

        return string

    def __repr__(self) -> str:
        npopt = np.get_printoptions()
        np.set_printoptions(threshold=5, edgeitems=2, linewidth=100)
        out = (
            self.__class__.__name__
            + "(dict="
            + dict.__repr__(self)
            + f", attrs={self.attrs!r})"
        )
        np.set_printoptions(**npopt)
        return " ".join(out.replace("\n", " ").split())

    def view_as(self) -> None:
        r"""View the Struct data as a third-party format data structure.

        Error
        -----
        Not implemented. Since Struct's fields can have different lengths,
        converting to a NumPy, Pandas or Awkward is generally not possible.
        Call :meth:`.LGDO.view_as` on the fields instead.

        See Also
        --------
        .LGDO.view_as
        """
        msg = (
            "Since Struct's fields can have different lengths, "
            "converting to a NumPy, Pandas or Awkward is generally "
            "not possible. Call view_as() on the fields instead."
        )
        raise NotImplementedError(msg)



from __future__ import annotations

import numpy as np
import pytest

from lgdo import lh5, types
from lgdo.lh5 import concat


def test_concat(lgnd_test_data, tmptestdir):
    infile1 = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_raw.lh5"
    )
    infile2 = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/phy/p03/r001/l200-p03-r001-phy-20230322T170202Z-tier_raw.lh5"
    )
    outfile = f"{tmptestdir}/out.lh5"
    concat.lh5concat(output=outfile, lh5_files=[infile1, infile2], overwrite=True)

    assert lh5.ls(outfile) == [
        "ch1057600",
        "ch1059201",
        "ch1062405",
        "ch1084803",
        "ch1084804",
        "ch1121600",
    ]
    assert lh5.ls(outfile, "ch1057600/raw/") == [
        "ch1057600/raw/abs_delta_mu_usec",
        "ch1057600/raw/baseline",
        "ch1057600/raw/board_id",
        "ch1057600/raw/channel",
        "ch1057600/raw/crate",
        "ch1057600/raw/daqenergy",
        "ch1057600/raw/deadtime",
        "ch1057600/raw/delta_mu_usec",
        "ch1057600/raw/dr_maxticks",
        "ch1057600/raw/dr_start_pps",
        "ch1057600/raw/dr_start_ticks",
        "ch1057600/raw/dr_stop_pps",
        "ch1057600/raw/dr_stop_ticks",
        "ch1057600/raw/event_type",
        "ch1057600/raw/eventnumber",
        "ch1057600/raw/fc_input",
        "ch1057600/raw/fcid",
        "ch1057600/raw/mu_offset_sec",
        "ch1057600/raw/mu_offset_usec",
        "ch1057600/raw/numtraces",
        "ch1057600/raw/packet_id",
        "ch1057600/raw/runtime",
        "ch1057600/raw/slot",
        "ch1057600/raw/timestamp",
        "ch1057600/raw/to_master_sec",
        "ch1057600/raw/to_start_sec",
        "ch1057600/raw/to_start_usec",
        "ch1057600/raw/tracelist",
        "ch1057600/raw/ts_maxticks",
        "ch1057600/raw/ts_pps",
        "ch1057600/raw/ts_ticks",
        "ch1057600/raw/waveform",
    ]
    assert lh5.ls(outfile, "ch1057600/raw/waveform/") == [
        "ch1057600/raw/waveform/dt",
        "ch1057600/raw/waveform/t0",
        "ch1057600/raw/waveform/values",
    ]

    store = lh5.LH5Store()
    tbl1 = store.read("ch1057600/raw", infile1)
    tbl2 = store.read("ch1057600/raw", infile2)
    tbl = store.read("ch1057600/raw", outfile)
    assert len(tbl) == 20

    for i in range(10):
        assert tbl.packet_id[i] == tbl1.packet_id[i]
        assert np.array_equal(tbl.tracelist[i], tbl1.tracelist[i])
        assert np.array_equal(tbl.waveform.values[i], tbl1.waveform.values[i])
    for i in range(10, 20):
        assert tbl.packet_id[i] == tbl2.packet_id[i - 10]
        assert np.array_equal(tbl.tracelist[i], tbl2.tracelist[i - 10])
        assert np.array_equal(tbl.waveform.values[i], tbl2.waveform.values[i - 10])

    concat.lh5concat(
        output=outfile,
        lh5_files=[infile1, infile2],
        overwrite=True,
        include_list=["ch1057600/raw/waveform/*"],
    )
    assert lh5.ls(outfile) == [
        "ch1057600",
    ]

    concat.lh5concat(
        output=outfile,
        lh5_files=[infile1, infile2],
        overwrite=True,
        include_list=["ch1057600/raw/waveform/values"],
    )

    assert lh5.ls(outfile, "ch1057600/raw/waveform/") == [
        "ch1057600/raw/waveform/values",
    ]

    tbl = store.read("ch1057600/raw", outfile)
    assert isinstance(tbl, types.Table)

    concat.lh5concat(
        output=outfile,
        lh5_files=[infile1, infile2],
        overwrite=True,
        exclude_list=["ch1057600/raw/waveform/values"],
    )
    assert lh5.ls(outfile) == [
        "ch1057600",
        "ch1059201",
        "ch1062405",
        "ch1084803",
        "ch1084804",
        "ch1121600",
    ]
    assert lh5.ls(outfile, "ch1059201/raw/waveform/") == [
        "ch1059201/raw/waveform/dt",
        "ch1059201/raw/waveform/t0",
        "ch1059201/raw/waveform/values",
    ]
    assert lh5.ls(outfile, "ch1057600/raw/waveform/") == [
        "ch1057600/raw/waveform/dt",
        "ch1057600/raw/waveform/t0",
    ]

    tbl1 = store.read("ch1059201/raw", infile1)
    tbl2 = store.read("ch1059201/raw", infile2)
    tbl = store.read("ch1059201/raw", outfile)
    assert len(tbl) == 20

    for i in range(10):
        assert tbl.packet_id[i] == tbl1.packet_id[i]
        assert np.array_equal(tbl.tracelist[i], tbl1.tracelist[i])
        assert np.array_equal(tbl.waveform.values[i], tbl1.waveform.values[i])
    for i in range(10, 20):
        assert tbl.packet_id[i] == tbl2.packet_id[i - 10]
        assert np.array_equal(tbl.tracelist[i], tbl2.tracelist[i - 10])
        assert np.array_equal(tbl.waveform.values[i], tbl2.waveform.values[i - 10])

    # test concatenating arrays in structs.
    infile1 = f"{tmptestdir}/concat_test_struct_0.lh5"
    tb1 = types.Table(col_dict={"col": types.Array(np.zeros(4))})
    struct1 = types.Struct({"x": tb1})
    store.write(struct1, "stp", infile1, wo_mode="overwrite_file")

    infile2 = f"{tmptestdir}/concat_test_struct_1.lh5"
    tb2 = types.Table(col_dict={"col": types.Array(np.ones(7))})
    struct2 = types.Struct({"x": tb2})
    store.write(struct2, "stp", infile2, wo_mode="overwrite_file")

    outfile = f"{tmptestdir}/concat_test_struct_out.lh5"
    concat.lh5concat(output=outfile, lh5_files=[infile1, infile2], overwrite=True)

    out_stp = store.read("stp", outfile)
    assert out_stp.attrs["datatype"] == "struct{x}"
    assert np.all(out_stp.x["col"].nda == np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1]))

    # check for reasonable failures
    with pytest.raises(RuntimeError):
        concat.lh5concat(output=outfile, lh5_files=[infile1], overwrite=True)



"""Variable-length code compression algorithms."""

from __future__ import annotations

import logging
from dataclasses import dataclass

import numba
import numpy as np
from numpy import int32, ubyte, uint32
from numpy.typing import NDArray

from .. import types as lgdo
from ..utils import numba_defaults_kwargs as nb_kwargs
from .base import WaveformCodec

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class ULEB128ZigZagDiff(WaveformCodec):
    """ZigZag [#WikiZZ]_ encoding followed by Unsigned Little Endian Base 128 (ULEB128) [#WikiULEB128]_ encoding of array differences.

    .. [#WikiZZ] https://wikipedia.org/wiki/Variable-length_quantity#Zigzag_encoding
    .. [#WikiULEB128] https://wikipedia.org/wiki/LEB128#Unsigned_LEB128
    """

    codec: str = "uleb128_zigzag_diff"


def encode(
    sig_in: NDArray | lgdo.VectorOfVectors | lgdo.ArrayOfEqualSizedArrays,
    sig_out: NDArray[ubyte] = None,
) -> (
    (NDArray[ubyte], NDArray[uint32])
    | lgdo.VectorOfEncodedVectors
    | lgdo.ArrayOfEncodedEqualSizedArrays
):
    """Compress digital signal(s) with a variable-length encoding of its derivative.

    Wraps :func:`uleb128_zigzag_diff_array_encode` and adds support for encoding
    LGDOs.

    Note
    ----
    If `sig_in` is a NumPy array, no resizing of `sig_out` is performed. Not
    even of the internally allocated one.

    Because of the current (hardware vectorized) implementation, providing a
    pre-allocated :class:`.VectorOfEncodedVectors` or
    :class:`.ArrayOfEncodedEqualSizedArrays` as `sig_out` is not possible.

    Parameters
    ----------
    sig_in
        array(s) holding the input signal(s).
    sig_out
        pre-allocated unsigned 8-bit integer array(s) for the compressed
        signal(s). If not provided, a new one will be allocated.

    Returns
    -------
    sig_out, nbytes | LGDO
        given pre-allocated `sig_out` structure or new structure of unsigned
        8-bit integers, plus the number of bytes (length) of the encoded
        signal. If `sig_in` is an :class:`.LGDO`, only a newly allocated
        :class:`.VectorOfEncodedVectors` or
        :class:`.ArrayOfEncodedEqualSizedArrays` is returned.

    See Also
    --------
    uleb128_zigzag_diff_array_encode
    """
    if isinstance(sig_in, np.ndarray):
        s = sig_in.shape
        if len(sig_in) == 0:
            return np.empty(s[:-1] + (0,), dtype=ubyte), np.empty(0, dtype=uint32)

        if sig_out is None:
            # the encoded signal is an array of bytes
            # pre-allocate ubyte (uint8) array with a generous (but safe) size
            max_b = int(np.ceil(np.iinfo(sig_in.dtype).bits / 16) * 5)
            # expand last dimension
            sig_out = np.empty(s[:-1] + (s[-1] * max_b,), dtype=ubyte)

        if sig_out.dtype != ubyte:
            msg = "sig_out must be of type ubyte"
            raise ValueError(msg)

        # nbytes has one dimension less (the last one)
        nbytes = np.empty(s[:-1], dtype=uint32)

        uleb128_zigzag_diff_array_encode(sig_in, sig_out, nbytes)

        # return without resizing
        return sig_out, nbytes

    if isinstance(sig_in, lgdo.VectorOfVectors):
        if sig_out is not None:
            log.warning(
                "a pre-allocated VectorOfEncodedVectors was given "
                "to hold an encoded ArrayOfEqualSizedArrays. "
                "This is not supported at the moment, so a new one "
                "will be allocated to replace it"
            )
        # convert VectorOfVectors to ArrayOfEqualSizedArrays so it can be
        # directly passed to the low-level encoding routine
        sig_out_nda, nbytes = encode(sig_in.to_aoesa(fill_val=0, preserve_dtype=True))

        # build the encoded LGDO
        encoded_data = lgdo.ArrayOfEqualSizedArrays(nda=sig_out_nda).to_vov(
            cumulative_length=np.cumsum(nbytes, dtype=uint32)
        )
        # decoded_size is an array, compute it by diff'ing the original VOV
        decoded_size = np.diff(sig_in.cumulative_length, prepend=uint32(0))

        return lgdo.VectorOfEncodedVectors(encoded_data, decoded_size)

    if isinstance(sig_in, lgdo.ArrayOfEqualSizedArrays):
        if sig_out:
            log.warning(
                "a pre-allocated VectorOfEncodedVectors was given "
                "to hold an encoded ArrayOfEqualSizedArrays. "
                "This is not supported at the moment, so a new one "
                "will be allocated to replace it"
            )

        # encode the internal numpy array
        sig_out_nda, nbytes = encode(sig_in.nda)

        # build the encoded LGDO
        encoded_data = lgdo.ArrayOfEqualSizedArrays(nda=sig_out_nda).to_vov(
            cumulative_length=np.cumsum(nbytes, dtype=uint32)
        )
        return lgdo.ArrayOfEncodedEqualSizedArrays(
            encoded_data, decoded_size=sig_in.nda.shape[1]
        )

    if isinstance(sig_in, lgdo.Array):
        # encode the internal numpy array
        sig_out_nda, nbytes = encode(sig_in.nda, sig_out)
        return lgdo.Array(sig_out_nda), nbytes

    msg = f"unsupported input signal type ({type(sig_in)})"
    raise ValueError(msg)


def decode(
    sig_in: (NDArray[ubyte], NDArray[uint32])
    | lgdo.VectorOfEncodedVectors
    | lgdo.ArrayOfEncodedEqualSizedArrays,
    sig_out: NDArray | lgdo.ArrayOfEqualSizedArrays = None,
) -> (NDArray, NDArray[uint32]) | lgdo.VectorOfVectors | lgdo.ArrayOfEqualSizedArrays:
    """Deompress digital signal(s) with a variable-length encoding of its derivative.

    Wraps :func:`uleb128_zigzag_diff_array_decode` and adds support for decoding
    LGDOs.

    Note
    ----
    If `sig_in` is a NumPy array, no resizing (along the last dimension) of
    `sig_out` to its actual length is performed. Not even of the internally
    allocated one. If a pre-allocated :class:`.ArrayOfEqualSizedArrays` is
    provided, it won't be resized too. The internally allocated
    :class:`.ArrayOfEqualSizedArrays` `sig_out` has instead always the correct
    size.

    Because of the current (hardware vectorized) implementation, providing a
    pre-allocated :class:`.VectorOfVectors` as `sig_out` is not possible.

    Parameters
    ----------
    sig_in
        array(s) holding the input, compressed signal(s). Output of
        :func:`.encode`.
    sig_out
        pre-allocated array(s) for the decompressed signal(s).  If not
        provided, will allocate a 32-bit integer array(s) structure.

    Returns
    -------
    sig_out, nbytes | LGDO
        given pre-allocated structure or new structure of 32-bit integers, plus
        the number of bytes (length) of the decoded signal.

    See Also
    --------
    uleb128_zigzag_diff_array_decode
    """
    # expect the output of encode()
    if isinstance(sig_in, tuple):
        if sig_out is None:
            # allocate output array of the same shape (generous)
            sig_out = np.empty_like(sig_in[0], dtype=int32)

        # siglen has one dimension less (the last)
        s = sig_in[0].shape
        siglen = np.empty(s[:-1], dtype=uint32)

        if len(sig_in[0]) == 0:
            return sig_out, siglen

        # call low-level routine
        uleb128_zigzag_diff_array_decode(sig_in[0], sig_in[1], sig_out, siglen)

        return sig_out, siglen

    if isinstance(sig_in, lgdo.ArrayOfEncodedEqualSizedArrays):
        if sig_out is None:
            # initialize output structure with decoded_size
            sig_out = lgdo.ArrayOfEqualSizedArrays(
                dims=(1, 1),
                shape=(len(sig_in), sig_in.decoded_size.value),
                dtype=int32,
                attrs=sig_in.getattrs(),
            )

        siglen = np.empty(len(sig_in), dtype=uint32)
        # save original encoded vector lengths
        nbytes = np.diff(sig_in.encoded_data.cumulative_length.nda, prepend=uint32(0))

        if len(sig_in) == 0:
            return sig_out

        # convert vector of vectors to array of equal sized arrays
        # can now decode on the 2D matrix together with number of bytes to read per row
        _, siglen = decode(
            (sig_in.encoded_data.to_aoesa(fill_val=0, preserve_dtype=True).nda, nbytes),
            sig_out if isinstance(sig_out, np.ndarray) else sig_out.nda,
        )

        # sanity check
        assert np.all(sig_in.decoded_size.value == siglen)

        return sig_out

    if isinstance(sig_in, lgdo.VectorOfEncodedVectors):
        if sig_out:
            log.warning(
                "a pre-allocated VectorOfVectors was given "
                "to hold an encoded VectorOfVectors. "
                "This is not supported at the moment, so a new one "
                "will be allocated to replace it"
            )

        siglen = np.empty(len(sig_in), dtype=uint32)
        # save original encoded vector lengths
        nbytes = np.diff(sig_in.encoded_data.cumulative_length.nda, prepend=uint32(0))

        # convert vector of vectors to array of equal sized arrays
        # can now decode on the 2D matrix together with number of bytes to read per row
        sig_out, siglen = decode(
            (sig_in.encoded_data.to_aoesa(fill_val=0, preserve_dtype=True).nda, nbytes)
        )

        # sanity check
        assert np.array_equal(sig_in.decoded_size, siglen)

        # converto to VOV before returning
        return sig_out.to_vov(np.cumsum(siglen, dtype=uint32))

    msg = "unsupported input signal type"
    raise ValueError(msg)


@numba.vectorize(
    ["uint64(int64)", "uint32(int32)", "uint16(int16)"],
    **nb_kwargs(nopython=True),
)
def zigzag_encode(x: int | NDArray[int]) -> int | NDArray[int]:
    """ZigZag-encode [#WikiZZ]_ signed integer numbers."""
    return (x >> 31) ^ (x << 1)


@numba.vectorize(
    ["int64(uint64)", "int32(uint32)", "int16(uint16)"],
    **nb_kwargs(nopython=True),
)
def zigzag_decode(x: int | NDArray[int]) -> int | NDArray[int]:
    """ZigZag-decode [#WikiZZ]_ signed integer numbers."""
    return (x >> 1) ^ -(x & 1)


@numba.jit(["uint32(int64, byte[:])"], **nb_kwargs(nopython=True))
def uleb128_encode(x: int, encx: NDArray[ubyte]) -> int:
    """Compute a variable-length representation of an unsigned integer.

    Implements the Unsigned Little Endian Base-128 encoding [#WikiULEB128]_.
    Only positive numbers are expected, as no *two's complement* is applied.

    Parameters
    ----------
    x
        the number to be encoded.
    encx
        the encoded varint as a NumPy array of bytes.

    Returns
    -------
    nbytes
        size of varint in bytes
    """
    i = 0
    bits = x & 0x7F
    x >>= 7
    while x:
        encx[i] = 0x80 | bits
        bits = x & 0x7F
        i += 1
        x >>= 7

    encx[i] = bits
    # return size of varint in bytes
    return i + 1


@numba.jit(["UniTuple(uint32, 2)(byte[:])"], **nb_kwargs(nopython=True))
def uleb128_decode(encx: NDArray[ubyte]) -> (int, int):
    """Decode a variable-length integer into an unsigned integer.

    Implements the Unsigned Little Endian Base-128 decoding [#WikiULEB128]_.
    Only encoded positive numbers are expected, as no *two's complement* is
    applied.

    Parameters
    ----------
    encx
        the encoded varint as a NumPy array of bytes.

    Returns
    -------
    x, nread
        the decoded value and the number of bytes read from the input array.
    """
    if len(encx) <= 0:
        empty_msg = "input bytes array is empty"
        raise ValueError(empty_msg)

    x = pos = uint32(0)
    for b in encx:
        x = x | ((b & 0x7F) << pos)
        if (b & 0x80) == 0:
            return (x, int(pos / 7 + 1))
        pos += 7

        if pos >= 64:
            overflow_msg = "overflow during decoding of varint encoded number"
            raise OverflowError(overflow_msg)

    msg = "malformed varint"
    raise RuntimeError(msg)


@numba.guvectorize(
    [
        "void(uint16[:], byte[:], uint32[:])",
        "void(uint32[:], byte[:], uint32[:])",
        "void(uint64[:], byte[:], uint32[:])",
        "void(int16[:], byte[:], uint32[:])",
        "void(int32[:], byte[:], uint32[:])",
        "void(int64[:], byte[:], uint32[:])",
    ],
    "(n),(m),()",
    **nb_kwargs(nopython=True),
)
def uleb128_zigzag_diff_array_encode(
    sig_in: NDArray[int], sig_out: NDArray[ubyte], nbytes: int
) -> None:
    """Encode an array of integer numbers.

    The algorithm computes the derivative (prepending 0 first) of `sig_in`,
    maps it to positive numbers by applying :func:`zigzag_encode` and finally
    computes its variable-length binary representation with
    :func:`uleb128_encode`.

    The encoded data is stored in `sig_out` as an array of bytes. The number of
    bytes written is stored in `nbytes`. The actual encoded data can therefore
    be found in ``sig_out[:nbytes]``.

    Parameters
    ----------
    sig_in
        the input array of integers.
    sig_out
        pre-allocated bytes array for the output encoded data.
    nbytes
        pre-allocated output array holding the number of bytes written (stored
        in the first index).

    See Also
    --------
    .uleb128_zigzag_diff_array_decode
    """
    pos = uint32(0)
    last = int32(0)
    for s in sig_in:
        zzdiff = zigzag_encode(int32(s - last))
        pos += uleb128_encode(zzdiff, sig_out[pos:])
        last = s

    nbytes[0] = pos


@numba.guvectorize(
    [
        "void(byte[:], uint32[:], uint16[:], uint32[:])",
        "void(byte[:], uint32[:], uint32[:], uint32[:])",
        "void(byte[:], uint32[:], uint64[:], uint32[:])",
        "void(byte[:], uint32[:], int16[:], uint32[:])",
        "void(byte[:], uint32[:], int32[:], uint32[:])",
        "void(byte[:], uint32[:], int64[:], uint32[:])",
    ],
    "(n),(),(m),()",
    **nb_kwargs(nopython=True),
)
def uleb128_zigzag_diff_array_decode(
    sig_in: NDArray[ubyte],
    nbytes: int,
    sig_out: NDArray[int],
    siglen: int,
) -> None:
    """Decode an array of variable-length integers.

    The algorithm inverts :func:`.uleb128_zigzag_diff_array_encode` by decoding
    the variable-length binary data in `sig_in` with :func:`uleb128_decode`,
    then reconstructing the original signal derivative with
    :func:`zigzag_decode` and finally computing its cumulative (i.e. the
    original signal).

    Parameters
    ----------
    sig_in
        the array of bytes encoding the variable-length integers.
    nbytes
        the number of bytes to read from `sig_in` (stored in the first index of
        this array).
    sig_out
        pre-allocated array for the output decoded signal.
    siglen
        the length of the decoded signal, (stored in the first index of this
        array).

    See Also
    --------
    .uleb128_zigzag_diff_array_encode
    """
    if len(sig_in) <= 0:
        msg = "input bytes array is empty"
        raise ValueError(msg)

    _nbytes = min(nbytes[0], len(sig_in))
    pos = i = uint32(0)
    last = int32(0)
    while pos < _nbytes:
        x, nread = uleb128_decode(sig_in[pos:])
        sig_out[i] = last = zigzag_decode(x) + last
        i += 1
        pos += nread

    siglen[0] = i



from __future__ import annotations

import bisect
import logging
import sys

import h5py
import numpy as np

from ....types import (
    Array,
    ArrayOfEncodedEqualSizedArrays,
    ArrayOfEqualSizedArrays,
    FixedSizeArray,
    Histogram,
    Scalar,
    Struct,
    Table,
    VectorOfEncodedVectors,
    VectorOfVectors,
    WaveformTable,
)
from ... import datatype as dtypeutils
from ...exceptions import LH5DecodeError
from . import utils
from .array import (
    _h5_read_array,
    _h5_read_array_of_equalsized_arrays,
    _h5_read_fixedsize_array,
)
from .encoded import (
    _h5_read_array_of_encoded_equalsized_arrays,
    _h5_read_vector_of_encoded_vectors,
)
from .scalar import _h5_read_scalar
from .vector_of_vectors import _h5_read_vector_of_vectors

log = logging.getLogger(__name__)


def _h5_read_lgdo(
    h5o,
    fname,
    oname,
    start_row=0,
    n_rows=sys.maxsize,
    idx=None,
    use_h5idx=False,
    field_mask=None,
    obj_buf=None,
    obj_buf_start=0,
    decompress=True,
):
    log.debug(
        f"reading {fname}:{oname}[{start_row}:{n_rows}], decompress = {decompress}, "
        + (f" with field mask {field_mask}" if field_mask else "")
    )

    attrs = utils.read_attrs(h5o, fname, oname)
    try:
        lgdotype = dtypeutils.datatype(attrs["datatype"])
    except KeyError as e:
        msg = "dataset not in file or missing 'datatype' attribute"
        raise LH5DecodeError(msg, fname, oname) from e

    if lgdotype is Scalar:
        return _h5_read_scalar(
            h5o,
            fname,
            oname,
            obj_buf=obj_buf,
        )

    # Convert whatever we input into a defaultdict
    field_mask = utils.build_field_mask(field_mask)

    if lgdotype is Struct:
        return _h5_read_struct(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            field_mask=field_mask,
            decompress=decompress,
        )

    # Below here is all array-like types. So trim idx if needed
    if idx is not None:
        # check if idx is just an ordered list of the integers if so can ignore
        if (idx == np.arange(0, len(idx), 1)).all():
            n_rows = min(n_rows, len(idx))
            idx = None
        else:
            # chop off indices < start_row
            i_first_valid = bisect.bisect_left(idx, start_row)
            idxa = idx[i_first_valid:]
            # don't readout more than n_rows indices
            idx = idxa[:n_rows]  # works even if n_rows > len(idxa)

    if lgdotype is Table:
        return _h5_read_table(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            field_mask=field_mask,
            obj_buf=obj_buf,
            obj_buf_start=obj_buf_start,
            decompress=decompress,
        )

    if lgdotype is Histogram:
        return _h5_read_histogram(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            field_mask=field_mask,
            obj_buf=obj_buf,
            obj_buf_start=obj_buf_start,
            decompress=decompress,
        )

    if lgdotype is ArrayOfEncodedEqualSizedArrays:
        return _h5_read_array_of_encoded_equalsized_arrays(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            obj_buf=obj_buf,
            obj_buf_start=obj_buf_start,
            decompress=decompress,
        )

    if lgdotype is VectorOfEncodedVectors:
        return _h5_read_vector_of_encoded_vectors(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            obj_buf=obj_buf,
            obj_buf_start=obj_buf_start,
            decompress=decompress,
        )

    if lgdotype is VectorOfVectors:
        return _h5_read_vector_of_vectors(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            obj_buf=obj_buf,
            obj_buf_start=obj_buf_start,
        )

    if lgdotype is FixedSizeArray:
        return _h5_read_fixedsize_array(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            obj_buf=obj_buf,
            obj_buf_start=obj_buf_start,
        )

    if lgdotype is ArrayOfEqualSizedArrays:
        return _h5_read_array_of_equalsized_arrays(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            obj_buf=obj_buf,
            obj_buf_start=obj_buf_start,
        )

    if lgdotype is Array:
        return _h5_read_array(
            h5o,
            fname,
            oname,
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            obj_buf=obj_buf,
            obj_buf_start=obj_buf_start,
        )

    msg = f"no rule to decode {lgdotype.__name__} from LH5"
    raise LH5DecodeError(msg, fname, oname)


def _h5_read_struct(
    h5g,
    fname,
    oname,
    start_row=0,
    n_rows=sys.maxsize,
    idx=None,
    use_h5idx=False,
    field_mask=None,
    decompress=True,
):
    # TODO: it's strange to pass start_row, n_rows, idx to struct
    # fields. If they all had shared indexing, they should be in a
    # table... Maybe should emit a warning? Or allow them to be
    # dicts keyed by field name?

    attrs = utils.read_attrs(h5g, fname, oname)

    # determine fields to be read out
    all_fields = dtypeutils.get_struct_fields(attrs["datatype"])
    selected_fields = utils.eval_field_mask(field_mask, all_fields)

    # modify datatype in attrs if a field_mask was used
    attrs["datatype"] = (
        "struct{" + ",".join(field for field, _ in selected_fields) + "}"
    )

    # loop over fields and read
    obj_dict = {}
    for field, submask in selected_fields:
        # support for integer keys
        field_key = int(field) if attrs.get("int_keys") else str(field)
        h5o = h5py.h5o.open(h5g, field.encode("utf-8"))
        obj_dict[field_key], _ = _h5_read_lgdo(
            h5o,
            fname,
            f"{oname}/{field}",
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            field_mask=submask,
            decompress=decompress,
        )
        h5o.close()

    return Struct(obj_dict=obj_dict, attrs=attrs), 1


def _h5_read_table(
    h5g,
    fname,
    oname,
    start_row=0,
    n_rows=sys.maxsize,
    idx=None,
    use_h5idx=False,
    field_mask=None,
    obj_buf=None,
    obj_buf_start=0,
    decompress=True,
):
    if obj_buf is not None and not isinstance(obj_buf, Table):
        msg = "provided object buffer is not a Table"
        raise LH5DecodeError(msg, fname, oname)

    attrs = utils.read_attrs(h5g, fname, oname)

    # determine fields to be read out
    all_fields = dtypeutils.get_struct_fields(attrs["datatype"])
    selected_fields = utils.eval_field_mask(field_mask, all_fields)

    # modify datatype in attrs if a field_mask was used
    attrs["datatype"] = "table{" + ",".join(field for field, _ in selected_fields) + "}"

    # read out each of the fields
    col_dict = {}
    rows_read = []
    for field, submask in selected_fields:
        fld_buf = None
        if obj_buf is not None:
            if not isinstance(obj_buf, Table) or field not in obj_buf:
                msg = "provided object buffer is not a Table or columns are missing"
                raise LH5DecodeError(msg, fname, oname)

            fld_buf = obj_buf[field]

        h5o = h5py.h5o.open(h5g, field.encode("utf-8"))
        col_dict[field], n_rows_read = _h5_read_lgdo(
            h5o,
            fname,
            f"{oname}/{field}",
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            obj_buf=fld_buf,
            obj_buf_start=obj_buf_start,
            field_mask=submask,
            decompress=decompress,
        )
        h5o.close()

        if obj_buf is not None and obj_buf_start + n_rows_read > len(obj_buf):
            obj_buf.resize(obj_buf_start + n_rows_read)

        rows_read.append(n_rows_read)

    # warn if all columns don't read in the same number of rows
    if len(rows_read) > 0:
        n_rows_read = rows_read[0]
    else:
        n_rows_read = 0
        log.warning(f"Table '{oname}' has no fields specified by {field_mask=}")

    for n in rows_read[1:]:
        if n != n_rows_read:
            log.warning(
                f"Table '{oname}' got strange n_rows_read = {n}, "
                "{n_rows_read} was expected ({rows_read})"
            )

    # fields have been read out, now return a table
    if obj_buf is None:
        # if col_dict contains just 3 objects called t0, dt, and values,
        # return a WaveformTable
        if (
            len(col_dict) == 3
            and "t0" in col_dict
            and "dt" in col_dict
            and "values" in col_dict
        ):
            table = WaveformTable(
                t0=col_dict["t0"], dt=col_dict["dt"], values=col_dict["values"]
            )
        else:
            table = Table(col_dict=col_dict, attrs=attrs)

        # set (write) loc to end of tree
        table.resize(do_warn=True)
        return table, n_rows_read

    # We have read all fields into the object buffer. Run
    # checks: All columns should be the same size. So update
    # table's size as necessary, warn if any mismatches are found
    obj_buf.resize(do_warn=True)

    # check attributes
    utils.check_obj_buf_attrs(obj_buf.attrs, attrs, fname, oname)

    return obj_buf, n_rows_read


def _h5_read_histogram(
    h5g,
    fname,
    oname,
    start_row=0,
    n_rows=sys.maxsize,
    idx=None,
    use_h5idx=False,
    field_mask=None,
    obj_buf=None,
    obj_buf_start=0,
    decompress=True,
):
    if obj_buf is not None or obj_buf_start != 0:
        msg = "reading a histogram into an existing object buffer is not supported"
        raise LH5DecodeError(msg, fname, oname)

    struct, n_rows_read = _h5_read_struct(
        h5g,
        fname,
        oname,
        start_row=start_row,
        n_rows=n_rows,
        idx=idx,
        use_h5idx=use_h5idx,
        field_mask=field_mask,
        decompress=decompress,
    )

    binning = []
    for _, a in struct.binning.items():
        be = a.binedges
        if isinstance(be, Struct):
            b = (None, be.first.value, be.last.value, be.step.value, a.closedleft.value)
        elif isinstance(be, Array):
            b = (be, None, None, None, a.closedleft.value)
        else:
            msg = "unexpected binning of histogram"
            raise LH5DecodeError(msg, fname, oname)
        ax = Histogram.Axis(*b)
        # copy attrs to "clone" the "whole" struct.
        ax.attrs = a.getattrs(datatype=True)
        ax["binedges"].attrs = be.getattrs(datatype=True)
        binning.append(ax)

    isdensity = struct.isdensity.value
    weights = struct.weights
    attrs = struct.getattrs(datatype=True)
    histogram = Histogram(weights, binning, isdensity, attrs=attrs)

    return histogram, n_rows_read



from __future__ import annotations

import h5py
import numpy as np

from lgdo import utils


def test_get_element_type():
    # variable length HD5 string datatype.
    h5py_str_dtype = h5py.string_dtype(encoding="ascii", length=None)

    objs = [
        ("hi", "string"),
        (True, "bool"),
        (np.void(0), "blob"),
        (0, "real"),
        (np.uint8(0), "real"),
        (float(0), "real"),
        (1 + 1j, "complex"),
        (b"hi", "string"),
        (np.array(["hi"]), "string"),
        (np.array([b"hi"], h5py_str_dtype), "string"),
    ]

    for obj, name in objs:
        get_name = utils.get_element_type(obj)
        assert get_name == name



from __future__ import annotations

import numpy as np

import lgdo
from lgdo import lh5


def test_read(lh5_file):
    lh5_obj = lh5.read("/data/struct/scalar", lh5_file)
    assert isinstance(lh5_obj, lgdo.Scalar)
    assert lh5_obj.value == 10
    assert lh5_obj.attrs["sth"] == 1


def test_write(tmptestdir):
    struct = lgdo.Struct()
    struct.add_field("scalar", lgdo.Scalar(value=10, attrs={"sth": 1}))
    lh5.write(
        struct,
        "struct",
        f"{tmptestdir}/tmp-pygama-lgdo-types2.lh5",
        group="/data",
        start_row=1,
        n_rows=3,
        wo_mode="overwrite_file",
    )
    assert lh5.ls(f"{tmptestdir}/tmp-pygama-lgdo-types2.lh5")


def test_read_as(lh5_file):
    store = lh5.LH5Store()
    obj1 = store.read("/data/struct/table", lh5_file, start_row=1)
    obj1 = obj1.view_as("pd", with_units=True)

    obj2 = lh5.read_as(
        "/data/struct/table", lh5_file, "pd", start_row=1, with_units=True
    )
    assert obj1.equals(obj2)

    obj2 = lh5.read_as("/data/struct/table", [lh5_file], "ak")


def test_read_multiple_files(lh5_file):
    lh5_obj = lh5.read("/data/struct/array", [lh5_file, lh5_file, lh5_file])
    assert isinstance(lh5_obj, lgdo.Array)
    assert len(lh5_obj) == 9
    assert (lh5_obj.nda == np.array([2, 3, 4] * 3)).all()

    lh5_obj = lh5.read(
        "/data/struct/array", [lh5_file, lh5_file, lh5_file], idx=[1, 3, 5, 7]
    )
    assert len(lh5_obj) == 4
    assert (lh5_obj.nda == np.array([3, 2, 4, 3])).all()



from __future__ import annotations

import re
from dataclasses import asdict, dataclass


@dataclass(frozen=True)
class WaveformCodec:
    """Base class identifying a waveform compression algorithm.

    The `self.codec` property returns a string identifier suitable for labeling
    encoded data on disk. This identifier is constant for all class instances.

    Note
    ----
    This is an abstract type. The user must provided a concrete subclass.
    """

    @property
    def codec(self):
        """The waveform codec string identifier.

        Will be attached as an attribute to the encoded Waveform values.
        """
        return re.sub("(?<!^)(?=[A-Z])", "_", type(self).__name__).lower()

    def asdict(self):
        """Return the dataclass fields as dictionary."""
        return {"codec": self.codec} | asdict(self)



"""Implements a LEGEND Data Object representing a scalar and corresponding utilities."""

from __future__ import annotations

import logging
from typing import Any

import numpy as np

from .. import utils
from ..units import default_units_registry as u
from .lgdo import LGDO

log = logging.getLogger(__name__)


class Scalar(LGDO):
    """Holds just a scalar value and some attributes (datatype, units, ...)."""

    # TODO: do scalars need proper numpy dtypes?

    def __init__(
        self, value: int | float | str, attrs: dict[str, Any] | None = None
    ) -> None:
        """
        Parameters
        ----------
        value
            the value for this scalar.
        attrs
            a set of user attributes to be carried along with this LGDO.
        """
        if not np.isscalar(value):
            msg = "cannot instantiate a Scalar with a non-scalar value"
            raise ValueError(msg)

        self.value = value
        super().__init__(attrs)

    def datatype_name(self) -> str:
        if hasattr(self.value, "datatype_name"):
            return self.value.datatype_name

        return utils.get_element_type(self.value)

    def form_datatype(self) -> str:
        return self.datatype_name()

    def view_as(self, with_units: bool = False):
        r"""Dummy function, returns the scalar value itself.

        See Also
        --------
        .LGDO.view_as
        """
        if with_units:
            return self.value * u[self.attrs["units"]]
        return self.value

    def __eq__(self, other: Scalar) -> bool:
        if isinstance(other, Scalar):
            return self.value == other.value and self.attrs == self.attrs

        return False

    def __str__(self) -> str:
        attrs = self.getattrs()
        return f"{self.value!s} with attrs={attrs!r}"

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"(value={self.value!r}, attrs={self.attrs!r})"



"""
Implements a LEGEND Data Object representing a special
:class:`~.lgdo.table.Table` to store blocks of one-dimensional time-series
data.
"""

from __future__ import annotations

import logging
from typing import Any

import awkward as ak
import numpy as np
import pandas as pd

from .array import Array
from .arrayofequalsizedarrays import ArrayOfEqualSizedArrays
from .encoded import ArrayOfEncodedEqualSizedArrays, VectorOfEncodedVectors
from .table import Table
from .vectorofvectors import VectorOfVectors

log = logging.getLogger(__name__)


class WaveformTable(Table):
    r"""An LGDO for storing blocks of (1D) time-series data.

    A :class:`WaveformTable` is an LGDO :class:`.Table` with the 3
    columns ``t0``, ``dt``, and ``values``:

    * ``t0[i]`` is a time offset (relative to a user-defined global reference)
      for the sample in ``values[i][0]``. Implemented as an LGDO
      :class:`.Array` with optional attribute ``units``.
    * ``dt[i]`` is the sampling period for the waveform at ``values[i]``.
      Implemented as an LGDO :class:`.Array` with optional attribute ``units``.
    * ``values[i]`` is the ``i``'th waveform in the table. Internally, the
      waveforms values may be either an LGDO :class:`.ArrayOfEqualSizedArrays`\
      ``<1,1>``, an LGDO :class:`.VectorOfVectors` or
      :class:`.VectorOfEncodedVectors` that supports waveforms of unequal
      length. Can optionally be given a ``units`` attribute.

    Note
    ----
    On-disk and in-memory versions could be different e.g. if a compression
    routine is used.
    """

    def __init__(
        self,
        size: int | None = None,
        t0: float | Array | np.ndarray = 0,
        t0_units: str | None = None,
        dt: float | Array | np.ndarray = 1,
        dt_units: str | None = None,
        values: ArrayOfEqualSizedArrays | VectorOfVectors | np.ndarray = None,
        values_units: str | None = None,
        wf_len: int | None = None,
        dtype: np.dtype = None,
        attrs: dict[str, Any] | None = None,
    ) -> None:
        r"""
        Parameters
        ----------
        size
            sets the number of rows in the table. If ``None``, the size will be
            determined from the first among `t0`, `dt`, or `values` to return a
            valid length. If not ``None``, `t0`, `dt`, and values will be
            resized as necessary to match `size`. If `size` is ``None`` and
            `t0`, `dt`, and `values` are all non-array-like, a default size of
            1024 is used.
        t0
            :math:`t_0` values to be used (or broadcast) to the `t0` column.
        t0_units
            units for the :math:`t_0` values. If not ``None`` and `t0` is an
            LGDO :class:`.Array`, overrides what's in `t0`.
        dt
            :math:`\delta t` values (sampling period) to be used (or
            broadcasted) to the `t0` column.
        dt_units
            units for the `dt` values. If not ``None`` and `dt` is an LGDO
            :class:`.Array`, overrides what's in `dt`.
        values
            The waveform data to be stored in the table. If ``None`` a block of
            data is prepared based on the `wf_len` and `dtype` arguments.
        values_units
            units for the waveform values. If not ``None`` and `values` is an
            LGDO :class:`.Array`, overrides what's in `values`.
        wf_len
            The length of the waveforms in each entry of a table. If ``None``
            (the default), unequal lengths are assumed and
            :class:`.VectorOfVectors` is used for the `values` column. Ignored
            if `values` is a 2D ndarray, in which case ``values.shape[1]`` is
            used.
        dtype
            The NumPy :class:`numpy.dtype` of the waveform data. If `values` is
            not ``None``, this argument is ignored. If both `values` and
            `dtype` are ``None``, :class:`numpy.float64` is used.
        attrs
            A set of user attributes to be carried along with this LGDO.
        """

        if size is None:
            if hasattr(t0, "__len__"):
                size = len(t0)
            elif hasattr(dt, "__len__"):
                size = len(dt)
            elif hasattr(values, "__len__"):
                size = len(values)
            if size is None:
                size = 1024

        if not isinstance(t0, Array):
            shape = (size,)
            t0_dtype = t0.dtype if hasattr(t0, "dtype") else np.float32
            if isinstance(t0, np.ndarray):
                t0 = Array(nda=t0, shape=shape, dtype=t0_dtype)
            else:
                t0 = Array(fill_val=t0, shape=shape, dtype=t0_dtype)

        if t0_units is not None:
            t0.attrs["units"] = f"{t0_units}"

        if not isinstance(dt, Array):
            shape = (size,)
            dt_dtype = dt.dtype if hasattr(dt, "dtype") else np.float32
            if isinstance(dt, np.ndarray):
                dt = Array(nda=dt, shape=shape, dtype=dt_dtype)
            else:
                dt = Array(fill_val=dt, shape=shape, dtype=dt_dtype)

        if dt_units is not None:
            dt.attrs["units"] = f"{dt_units}"

        if not isinstance(
            values,
            (
                ArrayOfEqualSizedArrays,
                VectorOfVectors,
                VectorOfEncodedVectors,
                ArrayOfEncodedEqualSizedArrays,
            ),
        ):
            if isinstance(values, np.ndarray):
                try:
                    wf_len = values.shape[1]
                except Exception:
                    wf_len = None
            if wf_len is None:  # make a VectorOfVectors
                shape_guess = (size, 100)
                if dtype is None:
                    dtype = np.dtype(np.float64)
                if values is None:
                    values = VectorOfVectors(shape_guess=shape_guess, dtype=dtype)
                else:
                    flattened_data = np.concatenate(values)
                    length = 0
                    cumulative_length = []
                    for i in range(size):
                        length += len(values[i])
                        cumulative_length.append(length)
                    values = VectorOfVectors(
                        flattened_data=flattened_data,
                        cumulative_length=cumulative_length,
                        dtype=dtype,
                    )
            else:  # make a ArrayOfEqualSizedArrays
                shape = (size, wf_len)
                if dtype is None:
                    dtype = (
                        values.dtype
                        if hasattr(values, "dtype")
                        else np.dtype(np.float64)
                    )
                if isinstance(values, np.ndarray):
                    values = ArrayOfEqualSizedArrays(
                        dims=(1, 1), nda=values, shape=shape, dtype=dtype
                    )
                else:
                    values = ArrayOfEqualSizedArrays(
                        dims=(1, 1), fill_val=0, shape=shape, dtype=dtype
                    )

        if values_units is not None:
            values.attrs["units"] = f"{values_units}"

        col_dict = {}
        col_dict["t0"] = t0
        col_dict["dt"] = dt
        col_dict["values"] = values
        super().__init__(size=size, col_dict=col_dict, attrs=attrs)

    @property
    def values(self) -> ArrayOfEqualSizedArrays | VectorOfVectors:
        return self["values"]

    @property
    def values_units(self) -> str:
        return self.values.attrs.get("units", None)

    @values_units.setter
    def values_units(self, units) -> None:
        self.values.attrs["units"] = f"{units}"

    @property
    def wf_len(self) -> int:
        if isinstance(self.values, VectorOfVectors):
            return -1
        return self.values.nda.shape[1]

    @wf_len.setter
    def wf_len(self, wf_len) -> None:
        if isinstance(self.values, VectorOfVectors):
            return
        shape = self.values.nda.shape
        shape = (shape[0], wf_len)
        self.values.resize(shape)

    def resize_wf_len(self, new_len: int) -> None:
        """Alias for `wf_len.setter`, for when we want to make it clear in
        the code that memory is being reallocated.
        """
        self.wf_len = new_len

    @property
    def t0(self) -> Array:
        return self["t0"]

    @property
    def t0_units(self) -> str:
        return self.t0.attrs.get("units", None)

    @t0_units.setter
    def t0_units(self, units: str) -> None:
        self.t0.attrs["units"] = f"{units}"

    @property
    def dt(self) -> Array:
        return self["dt"]

    @property
    def dt_units(self) -> str:
        return self.dt.attrs.get("units", None)

    @dt_units.setter
    def dt_units(self, units: str) -> None:
        self.dt.attrs["units"] = f"{units}"

    def __str__(self):
        npopt = np.get_printoptions()
        np.set_printoptions(threshold=100)

        string = ""

        for i in range(self.size):
            string += f"{self.values[i]}, dt={self.dt[i]}"
            if self.dt_units:
                string += f" {self.dt_units}"
            string += f", t0={self.t0[i]}"
            if self.t0_units:
                string += f" {self.t0_units}"
            if i < self.size - 1:
                string += "\n"

        np.set_printoptions(**npopt)
        return string

    def view_as(
        self,
        library: str,
        with_units: bool = False,
        cols: list[str] | None = None,
        prefix: str = "",
    ) -> pd.DataFrame | np.NDArray | ak.Array:
        r"""View the waveform data as a third-party format data structure.

        See Also
        --------
        .LGDO.view_as
        """
        return super().view_as(library, with_units, cols, prefix)



from __future__ import annotations

import numpy as np
import pytest

import lgdo
from lgdo import compression, lh5, types


@pytest.fixture(scope="module")
def lgnd_file(lgnd_test_data):
    return lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5")


@pytest.fixture(scope="module")
def lh5_file(tmptestdir):
    store = lh5.LH5Store()

    struct = lgdo.Struct()
    struct.add_field("scalar", lgdo.Scalar(value=10, attrs={"sth": 1}))
    struct.add_field("array", types.Array(nda=np.array([1, 2, 3, 4, 5])))
    struct.add_field("array2d", types.Array(shape=(23, 56), fill_val=69, dtype=int))
    struct.add_field(
        "aoesa",
        types.ArrayOfEqualSizedArrays(shape=(5, 5), dtype=np.float32, fill_val=42),
    )
    struct.add_field(
        "vov",
        types.VectorOfVectors(
            flattened_data=types.Array(
                nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1])
            ),
            cumulative_length=types.Array(nda=np.array([2, 5, 6, 10, 13])),
            attrs={"myattr": 2},
        ),
    )
    struct.add_field(
        "vov3d",
        types.VectorOfVectors(
            [[[1, 2], [3, 4, 5]], [[2], [4, 8, 9, 7]], [[5, 3, 1]], [[3], []], [[3, 4]]]
        ),
    )

    struct.add_field(
        "voev",
        types.VectorOfEncodedVectors(
            encoded_data=types.VectorOfVectors(
                flattened_data=types.Array(
                    nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1])
                ),
                cumulative_length=types.Array(nda=np.array([2, 5, 6, 10, 13])),
            ),
            decoded_size=types.Array(shape=5, fill_val=6),
        ),
    )

    col_dict = {
        "a": lgdo.Array(nda=np.array([1, 2, 3, 4]), attrs={"attr": 9}),
        "b": lgdo.Array(nda=np.array([5, 6, 7, 8]), attrs={"compression": "gzip"}),
        "c": lgdo.Array(
            nda=np.array([5, 6, 7, 8]),
            attrs={"compression": {"compression": "gzip", "compression_opts": 9}},
        ),
        "d": lgdo.Array(
            nda=np.array([5, 6, 7, 8]),
            attrs={"compression": None},
        ),
    }

    struct.add_field("table", types.Table(col_dict=col_dict, attrs={"stuff": 5}))

    struct.add_field(
        "wftable",
        types.WaveformTable(
            t0=types.Array(np.zeros(10)),
            dt=types.Array(np.full(10, fill_value=1)),
            values=types.ArrayOfEqualSizedArrays(
                shape=(10, 1000), dtype=np.uint16, fill_val=100, attrs={"custom": 8}
            ),
        ),
    )

    struct.add_field(
        "wftable_enc",
        types.WaveformTable(
            t0=types.Array(np.zeros(10)),
            dt=types.Array(np.full(10, fill_value=1)),
            values=compression.encode(
                struct["wftable"].values,
                codec=compression.RadwareSigcompress(codec_shift=-32768),
            ),
        ),
    )

    struct.add_field("empty_struct", types.Struct())

    store.write(
        struct,
        "struct",
        f"{tmptestdir}/tmp-pygama-lgdo-types.lh5",
        group="/data",
        start_row=1,
        n_rows=3,
        wo_mode="overwrite_file",
    )

    store.write(
        struct,
        "struct_full",
        f"{tmptestdir}/tmp-pygama-lgdo-types.lh5",
        group="/data",
        wo_mode="append",
    )

    assert struct["table"]["b"].attrs["compression"] == "gzip"

    return f"{tmptestdir}/tmp-pygama-lgdo-types.lh5"


@pytest.fixture(scope="module")
def enc_lgnd_file(lgnd_file, tmptestdir):
    store = lh5.LH5Store()
    wft = store.read("/geds/raw/waveform", lgnd_file)
    wft.values.attrs["compression"] = compression.RadwareSigcompress(codec_shift=-32768)
    store.write(
        wft,
        "/geds/raw/waveform",
        f"{tmptestdir}/tmp-pygama-compressed-wfs.lh5",
        wo_mode="overwrite_file",
    )
    return f"{tmptestdir}/tmp-pygama-compressed-wfs.lh5"



from __future__ import annotations

import fnmatch
import logging

from lgdo.lh5 import LH5Iterator

from .. import Array, Scalar, Struct, Table, VectorOfVectors, lh5

log = logging.getLogger(__name__)


def _get_obj_list(
    lh5_files: list, include_list: list | None = None, exclude_list: list | None = None
) -> list[str]:
    """Extract a list of lh5 objects to concatenate.

    Parameters
    ----------
    lh5_files
        list of input files to concatenate.
    include_list
        patterns for tables to include.
    exclude_list
        patterns for tables to exclude.

    """
    file0 = lh5_files[0]
    obj_list_full = set(lh5.ls(file0, recursive=True))

    # let's remove objects with nested LGDOs inside
    to_remove = set()
    for name in obj_list_full:
        if len(fnmatch.filter(obj_list_full, f"{name}/*")) > 1:
            to_remove.add(name)
    obj_list_full -= to_remove

    obj_list = set()
    # now first remove excluded stuff
    if exclude_list is not None:
        for exc in exclude_list:
            obj_list_full -= set(fnmatch.filter(obj_list_full, exc.strip("/")))

    # then make list of included, based on latest list
    if include_list is not None:
        for inc in include_list:
            obj_list |= set(fnmatch.filter(obj_list_full, inc.strip("/")))
    else:
        obj_list = obj_list_full

    # sort
    return sorted(obj_list)


def _get_lgdos(file, obj_list):
    """Get name of LGDO objects."""

    store = lh5.LH5Store()
    h5f0 = store.gimme_file(file)

    lgdos = []
    lgdo_structs = {}

    # loop over object list in the first file
    for name in obj_list:
        # now loop over groups starting from root
        current = ""
        for item in name.split("/"):
            current = f"{current}/{item}".strip("/")

            if current in lgdos:
                break

            # not even an LGDO (i.e. a plain HDF5 group)!
            if "datatype" not in h5f0[current].attrs:
                continue

            # read as little as possible
            obj = store.read(current, h5f0, n_rows=1)
            if isinstance(obj, (Table, Array, VectorOfVectors)):
                lgdos.append(current)

            elif isinstance(obj, Struct):
                # structs might be used in a "group-like" fashion (i.e. they might only
                # contain array-like objects).
                # note: handle after handling tables, as tables also satisfy this check.
                lgdo_structs[current] = obj.attrs["datatype"]
                continue

            elif isinstance(obj, Scalar):
                msg = f"cannot concat scalar field {current}"
                log.warning(msg)

            break

    msg = f"first-level, array-like objects: {lgdos}"
    log.info(msg)

    msg = f"nested structs: {lgdo_structs}"
    log.info(msg)

    h5f0.close()

    if lgdos == []:
        msg = "did not find any field to concatenate, exit"
        raise RuntimeError(msg)

    return lgdos, lgdo_structs


def _inplace_table_filter(name, table, obj_list):
    """filter objects nested in this LGDO"""
    skm = fnmatch.filter(obj_list, f"{name}/*")
    kept = {it.removeprefix(name).strip("/").split("/")[0] for it in skm}

    # now remove fields
    for k in list(table.keys()):
        if k not in kept:
            table.remove_column(k)

    msg = f"fields left in table '{name}': {table.keys()}"
    log.debug(msg)

    # recurse!
    for k2, v2 in table.items():
        if not isinstance(v2, Table):
            continue

        _inplace_table_filter(f"{name}/{k2}", v2, obj_list)


def _remove_nested_fields(lgdos: dict, obj_list: list):
    """Remove (nested) table fields based on obj_list."""

    for key, val in lgdos.items():
        if not isinstance(val, Table):
            continue

        _inplace_table_filter(key, val, obj_list)


def lh5concat(
    lh5_files: list,
    output: str,
    overwrite: bool = False,
    *,
    include_list: list | None = None,
    exclude_list: list | None = None,
) -> None:
    """Concatenate LGDO Arrays, VectorOfVectors and Tables in LH5 files.

    Parameters
    ----------
    lh5_files
        list of input files to concatenate.
    output
        path to the output file
    include_list
        patterns for tables to include.
    exclude_list
        patterns for tables to exclude.
    """

    if len(lh5_files) < 2:
        msg = "you must provide at least two input files"
        raise RuntimeError(msg)

    # determine list of objects by recursively ls'ing first file
    obj_list = _get_obj_list(
        lh5_files, include_list=include_list, exclude_list=exclude_list
    )

    msg = f"objects matching include patterns {include_list} in {lh5_files[0]}: {obj_list}"
    log.info(msg)

    lgdos, lgdo_structs = _get_lgdos(lh5_files[0], obj_list)
    first_done = False
    store = lh5.LH5Store()

    # loop over lgdo objects
    for lgdo in lgdos:
        # iterate over the files
        for lh5_obj in LH5Iterator(lh5_files, lgdo):
            data = {lgdo: lh5_obj}

            # remove the nested fields
            _remove_nested_fields(data, obj_list)

            if first_done is False:
                msg = f"creating output file {output}"
                log.info(msg)

                store.write(
                    data[lgdo],
                    lgdo,
                    output,
                    wo_mode="overwrite_file"
                    if (overwrite and not first_done)
                    else "write_safe",
                )
                first_done = True

            else:
                msg = f"appending to {output}"
                log.info(msg)

                if isinstance(data[lgdo], Table):
                    _inplace_table_filter(lgdo, data[lgdo], obj_list)

                store.write(data[lgdo], lgdo, output, wo_mode="append")

    if lgdo_structs != {}:
        output_file = store.gimme_file(output, mode="a")
        for struct, struct_dtype in lgdo_structs.items():
            msg = f"reset datatype of struct {struct} to {struct_dtype}"
            log.debug(msg)

            output_file[struct].attrs["datatype"] = struct_dtype
        output_file.close()



from __future__ import annotations

import h5py


class LH5DecodeError(Exception):
    def __init__(self, message: str, fname: str, oname: str) -> None:
        super().__init__(message)

        self.file = fname
        self.obj = oname

    def __str__(self) -> str:
        return (
            f"while reading object '{self.obj}' in file {self.file}: "
            + super().__str__()
        )

    def __reduce__(self) -> tuple:  # for pickling.
        return self.__class__, (*self.args, self.file, self.obj)


class LH5EncodeError(Exception):
    def __init__(
        self, message: str, file: str | h5py.File, group: str | h5py.Group, name: str
    ) -> None:
        super().__init__(message)

        self.file = file.filename if isinstance(file, h5py.File) else file
        self.group = (group.name if isinstance(file, h5py.File) else group).rstrip("/")
        self.name = name.lstrip("/")

    def __str__(self) -> str:
        return (
            f"while writing object {self.group}/{self.name} to file {self.file}: "
            + super().__str__()
        )

    def __reduce__(self) -> tuple:  # for pickling.
        return self.__class__, (*self.args, self.file, self.group, self.name)



from __future__ import annotations

import logging
import sys

import h5py

from .... import compression as compress
from ....types import (
    ArrayOfEncodedEqualSizedArrays,
    VectorOfEncodedVectors,
)
from ...exceptions import LH5DecodeError
from .array import (
    _h5_read_array,
)
from .scalar import _h5_read_scalar
from .utils import read_attrs
from .vector_of_vectors import _h5_read_vector_of_vectors

log = logging.getLogger(__name__)


def _h5_read_array_of_encoded_equalsized_arrays(
    h5g,
    fname,
    oname,
    **kwargs,
):
    return _h5_read_encoded_array(
        ArrayOfEncodedEqualSizedArrays, h5g, fname, oname, **kwargs
    )


def _h5_read_vector_of_encoded_vectors(
    h5g,
    fname,
    oname,
    **kwargs,
):
    return _h5_read_encoded_array(VectorOfEncodedVectors, h5g, fname, oname, **kwargs)


def _h5_read_encoded_array(
    lgdotype,
    h5g,
    fname,
    oname,
    start_row=0,
    n_rows=sys.maxsize,
    idx=None,
    use_h5idx=False,
    obj_buf=None,
    obj_buf_start=0,
    decompress=True,
):
    if lgdotype not in (ArrayOfEncodedEqualSizedArrays, VectorOfEncodedVectors):
        msg = f"unsupported read of encoded type {lgdotype.__name__}"
        raise LH5DecodeError(msg, fname, oname)

    if not decompress and obj_buf is not None and not isinstance(obj_buf, lgdotype):
        msg = f"object buffer is not a {lgdotype.__name__}"
        raise LH5DecodeError(msg, fname, oname)

    # read out decoded_size, either a Scalar or an Array
    decoded_size_buf = encoded_data_buf = None
    if obj_buf is not None and not decompress:
        decoded_size_buf = obj_buf.decoded_size
        encoded_data_buf = obj_buf.encoded_data

    if lgdotype is VectorOfEncodedVectors:
        h5o = h5py.h5o.open(h5g, b"decoded_size")
        decoded_size, _ = _h5_read_array(
            h5o,
            fname,
            f"{oname}/decoded_size",
            start_row=start_row,
            n_rows=n_rows,
            idx=idx,
            use_h5idx=use_h5idx,
            obj_buf=None if decompress else decoded_size_buf,
            obj_buf_start=0 if decompress else obj_buf_start,
        )
        h5o.close()
    else:
        h5o = h5py.h5o.open(h5g, b"decoded_size")
        decoded_size, _ = _h5_read_scalar(
            h5o,
            fname,
            f"{oname}/decoded_size",
            obj_buf=None if decompress else decoded_size_buf,
        )
        h5o.close()

    # read out encoded_data, a VectorOfVectors
    h5o = h5py.h5o.open(h5g, b"encoded_data")
    encoded_data, n_rows_read = _h5_read_vector_of_vectors(
        h5o,
        fname,
        f"{oname}/encoded_data",
        start_row=start_row,
        n_rows=n_rows,
        idx=idx,
        use_h5idx=use_h5idx,
        obj_buf=None if decompress else encoded_data_buf,
        obj_buf_start=0 if decompress else obj_buf_start,
    )
    h5o.close()

    # return the still encoded data in the buffer object, if there
    if obj_buf is not None and not decompress:
        return obj_buf, n_rows_read

    # otherwise re-create the encoded LGDO
    rawdata = lgdotype(
        encoded_data=encoded_data,
        decoded_size=decoded_size,
        attrs=read_attrs(h5g, fname, oname),
    )

    # already return if no decompression is requested
    if not decompress:
        return rawdata, n_rows_read

    # if no buffer, decode and return
    if obj_buf is None and decompress:
        return compress.decode(rawdata), n_rows_read

    # eventually expand provided obj_buf, if too short
    buf_size = obj_buf_start + n_rows_read
    if len(obj_buf) < buf_size:
        obj_buf.resize(buf_size)

    # use the (decoded object type) buffer otherwise
    if lgdotype is ArrayOfEncodedEqualSizedArrays:
        compress.decode(rawdata, obj_buf[obj_buf_start:buf_size])

    elif lgdotype is VectorOfEncodedVectors:
        # FIXME: not a good idea. an in place decoding version
        # of decode would be needed to avoid extra memory
        # allocations
        for i, wf in enumerate(compress.decode(rawdata)):
            obj_buf[obj_buf_start + i] = wf

    return obj_buf, n_rows_read



"""Implements utilities for LEGEND Data Objects."""

from __future__ import annotations

import glob
import logging
import os
import string
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any

import h5py

from .. import types
from . import _serializers
from .exceptions import LH5DecodeError

log = logging.getLogger(__name__)


def get_buffer(
    name: str,
    lh5_file: str | h5py.File | Sequence[str | h5py.File],
    size: int | None = None,
    field_mask: Mapping[str, bool] | Sequence[str] | None = None,
) -> types.LGDO:
    """Returns an LGDO appropriate for use as a pre-allocated buffer.

    Sets size to `size` if object has a size.
    """
    obj, n_rows = _serializers._h5_read_lgdo(
        lh5_file[name], n_rows=0, field_mask=field_mask
    )

    if hasattr(obj, "resize") and size is not None:
        obj.resize(new_size=size)

    return obj


def read_n_rows(name: str, h5f: str | h5py.File) -> int | None:
    """Look up the number of rows in an Array-like LGDO object on disk.

    Return ``None`` if `name` is a :class:`.Scalar` or a :class:`.Struct`.
    """
    if not isinstance(h5f, h5py.File):
        h5f = h5py.File(h5f, "r", locking=False)

    try:
        h5o = h5f[name].id
    except KeyError as e:
        msg = "not found"
        raise LH5DecodeError(msg, h5f, name) from e

    return _serializers.read.utils.read_n_rows(h5o, h5f.name, name)


def read_size_in_bytes(name: str, h5f: str | h5py.File) -> int | None:
    """Look up the size (in B) in an LGDO object in memory. Will crawl
    recursively through members of a Struct or Table
    """
    if not isinstance(h5f, h5py.File):
        h5f = h5py.File(h5f, "r", locking=False)

    try:
        h5o = h5f[name].id
    except KeyError as e:
        msg = "not found"
        raise LH5DecodeError(msg, h5f, name) from e

    return _serializers.read.utils.read_size_in_bytes(h5o, h5f.name, name)


def get_h5_group(
    group: str | h5py.Group,
    base_group: h5py.Group,
    grp_attrs: Mapping[str, Any] | None = None,
    overwrite: bool = False,
) -> h5py.Group:
    """
    Returns an existing :mod:`h5py` group from a base group or creates a
    new one. Can also set (or replace) group attributes.

    Parameters
    ----------
    group
        name of the HDF5 group.
    base_group
        HDF5 group to be used as a base.
    grp_attrs
        HDF5 group attributes.
    overwrite
        whether overwrite group attributes, ignored if `grp_attrs` is
        ``None``.
    """
    if not isinstance(group, h5py.Group):
        if group in base_group:
            group = base_group[group]
        else:
            group = base_group.create_group(group)
            if grp_attrs is not None:
                group.attrs.update(
                    {
                        k: v.encode("utf-8") if isinstance(v, str) else v
                        for k, v in grp_attrs.items()
                    }
                )
            return group
    if (
        grp_attrs is not None
        and len(set(grp_attrs.items()) ^ set(group.attrs.items())) > 0
    ):
        if not overwrite:
            msg = (
                f"Provided {grp_attrs=} are different from "
                f"existing ones {dict(group.attrs)=} but overwrite flag is not set"
            )
            raise RuntimeError(msg)

        log.debug(f"overwriting {group}.attrs...")
        for key in group.attrs:
            group.attrs.pop(key)

        group.attrs.update(
            {
                k: v.encode("utf-8") if isinstance(v, str) else v
                for k, v in grp_attrs.items()
            }
        )

    return group


def expand_vars(expr: str, substitute: dict[str, str] | None = None) -> str:
    """Expand (environment) variables.

    Note
    ----
    Malformed variable names and references to non-existing variables are left
    unchanged.

    Parameters
    ----------
    expr
        string expression, which may include (environment) variables prefixed by
        ``$``.
    substitute
        use this dictionary to substitute variables. Takes precedence over
        environment variables.
    """
    if substitute is None:
        substitute = {}

    # use provided mapping
    # then expand env variables
    return os.path.expandvars(string.Template(str(expr)).safe_substitute(substitute))


def expand_path(
    path: str,
    substitute: dict[str, str] | None = None,
    list: bool = False,
    base_path: str | None = None,
) -> str | list:
    """Expand (environment) variables and wildcards to return absolute paths.

    Parameters
    ----------
    path
        name of path, which may include environment variables and wildcards.
    list
        if ``True``, return a list. If ``False``, return a string; if ``False``
        and a unique file is not found, raise an exception.
    substitute
        use this dictionary to substitute variables. Environment variables take
        precedence.
    base_path
        name of base path. Returned paths will be relative to base.

    Returns
    -------
    path or list of paths
        Unique absolute path, or list of all absolute paths
    """
    if base_path is not None and base_path != "":
        base_path = Path(os.path.expandvars(base_path)).expanduser()
        path = base_path / path

    # first expand variables
    _path = expand_vars(path, substitute)

    # then expand wildcards
    # pathlib glob works differently so use glob for now
    paths = sorted(glob.glob(str(Path(_path).expanduser())))  # noqa: PTH207

    if base_path is not None and base_path != "":
        paths = [os.path.relpath(p, base_path) for p in paths]

    if not list:
        if len(paths) == 0:
            msg = f"could not find path matching {path}"
            raise FileNotFoundError(msg)
        if len(paths) > 1:
            msg = f"found multiple paths matching {path}"
            raise FileNotFoundError(msg)

        return paths[0]

    return paths


# https://stackoverflow.com/a/1094933
def fmtbytes(num, suffix="B"):
    """Returns formatted f-string for printing human-readable number of bytes."""
    for unit in ("", "k", "M", "G", "T", "P", "E", "Z"):
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f} Y{suffix}"



from __future__ import annotations

import logging
from pathlib import Path

import awkward as ak
import h5py
import numpy as np
import pytest

from lgdo import lh5, types


def test_write_compressed_lgnd_waveform_table(enc_lgnd_file):
    pass


def test_write_with_hdf5_compression(lgnd_file, tmptestdir):
    store = lh5.LH5Store()
    wft = store.read("/geds/raw/waveform", lgnd_file)
    store.write(
        wft,
        "/geds/raw/waveform",
        f"{tmptestdir}/tmp-pygama-hdf5-compressed-wfs.lh5",
        wo_mode="overwrite_file",
        compression="gzip",
        compression_opts=9,
        shuffle=True,
    )
    with h5py.File(f"{tmptestdir}/tmp-pygama-hdf5-compressed-wfs.lh5") as h5f:
        assert h5f["/geds/raw/waveform/values"].compression == "gzip"
        assert h5f["/geds/raw/waveform/values"].compression_opts == 9
        assert h5f["/geds/raw/waveform/values"].shuffle is True

    store.write(
        wft,
        "/geds/raw/waveform",
        f"{tmptestdir}/tmp-pygama-hdf5-compressed-wfs.lh5",
        wo_mode="overwrite_file",
        compression=None,
        shuffle=False,
    )
    with h5py.File(f"{tmptestdir}/tmp-pygama-hdf5-compressed-wfs.lh5") as h5f:
        assert h5f["/geds/raw/waveform/values"].compression is None
        assert h5f["/geds/raw/waveform/values"].shuffle is False

    store.write(
        wft.values,
        "/geds/raw/waveform/values",
        f"{tmptestdir}/tmp-pygama-hdf5-compressed-wfs.lh5",
        wo_mode="overwrite_file",
        chunks=[1, 10],
        compression=None,
        shuffle=False,
    )
    with h5py.File(f"{tmptestdir}/tmp-pygama-hdf5-compressed-wfs.lh5") as h5f:
        assert h5f["/geds/raw/waveform/values"].compression is None
        assert h5f["/geds/raw/waveform/values"].shuffle is False


def test_write_empty_vov(tmptestdir):
    vov = types.VectorOfVectors(flattened_data=[], cumulative_length=[])
    store = lh5.LH5Store()
    store.write(
        vov,
        "vov",
        f"{tmptestdir}/tmp-pygama-lgdo-empty-vov.lh5",
        group="/data",
    )

    obj = store.read("/data/vov", f"{tmptestdir}/tmp-pygama-lgdo-empty-vov.lh5")
    assert obj == vov


def test_write_append_array(tmptestdir):
    arr = types.Array([1, 2, 3, 4])
    arr_bis = types.Array([11, 34, 55, 57, 16])

    outfile = f"{tmptestdir}/write_append_array.lh5"
    lh5.write(arr, "arr", outfile, wo_mode="of")
    lh5.write(arr_bis, "arr", outfile, wo_mode="append")

    v = lh5.read("arr", outfile)
    assert v == types.Array([1, 2, 3, 4, 11, 34, 55, 57, 16])


def test_write_append_vov(tmptestdir):
    vov = types.VectorOfVectors([[1, 2, 3], [4], [5, 6], [7, 9, 8]])
    vov_bis = types.VectorOfVectors([[11], [34, 55], [57, 16], [28]])

    outfile = f"{tmptestdir}/write_append_array.lh5"
    lh5.write(vov, "vov", outfile, wo_mode="of")
    lh5.write(vov_bis, "vov", outfile, wo_mode="append")

    v = lh5.read("vov", outfile)
    assert ak.is_valid(v.view_as("ak"))
    assert v == types.VectorOfVectors(
        [[1, 2, 3], [4], [5, 6], [7, 9, 8], [11], [34, 55], [57, 16], [28]]
    )

    vov = types.VectorOfVectors([[[1, 2, 3], [4]], [[5, 6], [7, 9, 8]]])
    vov_bis = types.VectorOfVectors([[[11], [34, 55]], [[57, 16], [28]]])

    outfile = f"{tmptestdir}/write_append_array.lh5"
    lh5.write(vov, "vov", outfile, wo_mode="of")
    lh5.write(vov_bis, "vov", outfile, wo_mode="append")

    v = lh5.read("vov", outfile)
    assert ak.is_valid(v.view_as("ak"))


# First test that we can overwrite a table with the same name without deleting the original field
def test_write_object_overwrite_table_no_deletion(caplog, tmptestdir):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    if Path(f"{tmptestdir}/write_object_overwrite_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_overwrite_test.lh5").unlink()

    tb1 = types.Table(col_dict={"dset1": types.Array(np.zeros(10))})
    tb2 = types.Table(
        col_dict={"dset1": types.Array(np.ones(10))}
    )  # Same field name, different values
    store = lh5.LH5Store()
    store.write(tb1, "my_group", f"{tmptestdir}/write_object_overwrite_test.lh5")
    store.write(
        tb2,
        "my_group",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
    )  # Now, try to overwrite the same field

    # If the old field is deleted from the file before writing the new field, then we would get an extra debug statement
    assert "dset1 is not present in new table, deleting field" not in [
        rec.message for rec in caplog.records
    ]

    # Now, check that the data were overwritten
    tb_dat = store.read("my_group", f"{tmptestdir}/write_object_overwrite_test.lh5")
    assert np.array_equal(tb_dat["dset1"].nda, np.ones(10))


# Second: test that when we overwrite a table with fields with a different name, we delete the original field
def test_write_object_overwrite_table_with_deletion(caplog, tmptestdir):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    if Path(f"{tmptestdir}/write_object_overwrite_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_overwrite_test.lh5").unlink()

    tb1 = types.Table(col_dict={"dset1": types.Array(np.zeros(10))})
    tb2 = types.Table(
        col_dict={"dset2": types.Array(np.ones(10))}
    )  # Same field name, different values
    store = lh5.LH5Store()
    store.write(tb1, "my_group", f"{tmptestdir}/write_object_overwrite_test.lh5")
    store.write(
        tb2,
        "my_group",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
    )  # Now, try to overwrite with a different field

    # Now, check that the data were overwritten
    tb_dat = store.read("my_group", f"{tmptestdir}/write_object_overwrite_test.lh5")
    assert np.array_equal(tb_dat["dset2"].nda, np.ones(10))

    # Also make sure that the first table's fields aren't lurking around the lh5 file!
    with h5py.File(f"{tmptestdir}/write_object_overwrite_test.lh5", "r") as lh5file:
        assert "dset1" not in list(lh5file["my_group"].keys())

    # Make sure the same behavior happens when we nest the table in a group
    if Path(f"{tmptestdir}/write_object_overwrite_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_overwrite_test.lh5").unlink()

    tb1 = types.Table(col_dict={"dset1": types.Array(np.zeros(10))})
    tb2 = types.Table(
        col_dict={"dset2": types.Array(np.ones(10))}
    )  # Same field name, different values
    store = lh5.LH5Store()
    store.write(
        tb1,
        "my_table",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        group="my_group",
    )
    store.write(
        tb2,
        "my_table",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        group="my_group",
        wo_mode="overwrite",
    )  # Now, try to overwrite with a different field

    # Now, check that the data were overwritten
    tb_dat = store.read(
        "my_group/my_table", f"{tmptestdir}/write_object_overwrite_test.lh5"
    )
    assert np.array_equal(tb_dat["dset2"].nda, np.ones(10))

    # Also make sure that the first table's fields aren't lurking around the lh5 file!
    with h5py.File(f"{tmptestdir}/write_object_overwrite_test.lh5", "r") as lh5file:
        assert "dset1" not in list(lh5file["my_group/my_table"].keys())


# Third: test that when we overwrite other LGDO classes
def test_write_object_overwrite_lgdo(caplog, tmptestdir):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    # Start with an types.WaveformTable
    if Path(f"{tmptestdir}/write_object_overwrite_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_overwrite_test.lh5").unlink()

    tb1 = types.WaveformTable(
        t0=np.zeros(10),
        t0_units="ns",
        dt=np.zeros(10),
        dt_units="ns",
        values=np.zeros((10, 10)),
        values_units="ADC",
    )
    tb2 = types.WaveformTable(
        t0=np.ones(10),
        t0_units="ns",
        dt=np.ones(10),
        dt_units="ns",
        values=np.ones((10, 10)),
        values_units="ADC",
    )  # Same field name, different values
    store = lh5.LH5Store()
    store.write(
        tb1,
        "my_table",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        group="my_group",
    )
    store.write(
        tb2,
        "my_table",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
        group="my_group",
    )

    # If the old field is deleted from the file before writing the new field, then we would get a debug statement
    assert "my_table is not present in new table, deleting field" not in [
        rec.message for rec in caplog.records
    ]

    # Now, check that the data were overwritten
    tb_dat = store.read(
        "my_group/my_table", f"{tmptestdir}/write_object_overwrite_test.lh5"
    )
    assert np.array_equal(tb_dat["values"].nda, np.ones((10, 10)))

    # Now try overwriting an array, and test the write_start argument
    array1 = types.Array(nda=np.zeros(10))
    array2 = types.Array(nda=np.ones(20))
    store = lh5.LH5Store()
    store.write(array1, "my_array", f"{tmptestdir}/write_object_overwrite_test.lh5")
    store.write(
        array2,
        "my_array",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
        write_start=5,
    )

    # Now, check that the data were overwritten
    array_dat = store.read("my_array", f"{tmptestdir}/write_object_overwrite_test.lh5")
    expected_out_array = np.append(np.zeros(5), np.ones(20))

    assert np.array_equal(array_dat.nda, expected_out_array)

    # Now try overwriting a scalar
    scalar1 = types.Scalar(0)
    scalar2 = types.Scalar(1)
    store = lh5.LH5Store()
    store.write(scalar1, "my_scalar", f"{tmptestdir}/write_object_overwrite_test.lh5")
    store.write(
        scalar2,
        "my_scalar",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
    )

    # Now, check that the data were overwritten
    scalar_dat = store.read(
        "my_scalar", f"{tmptestdir}/write_object_overwrite_test.lh5"
    )

    assert scalar_dat.value == 1

    # Finally, try overwriting a vector of vectors
    vov1 = types.VectorOfVectors([np.zeros(1), np.ones(2), np.zeros(3)])
    vov2 = types.VectorOfVectors([np.ones(1), np.zeros(2), np.ones(3)])
    store = lh5.LH5Store()
    store.write(vov1, "my_vector", f"{tmptestdir}/write_object_overwrite_test.lh5")
    store.write(
        vov2,
        "my_vector",
        f"{tmptestdir}/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
        write_start=1,
    )  # start overwriting the second list of lists

    vector_dat = store.read(
        "my_vector", f"{tmptestdir}/write_object_overwrite_test.lh5"
    )

    assert np.array_equal(vector_dat.cumulative_length.nda, [1, 2, 4, 7])
    assert np.array_equal(vector_dat.flattened_data.nda, [0, 1, 0, 0, 1, 1, 1])


# Test that when we try to overwrite an existing column in a table we fail
def test_write_object_append_column(tmptestdir):
    # Try to append an array to a table
    if Path(f"{tmptestdir}/write_object_append_column_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_append_column_test.lh5").unlink()

    array1 = types.Array(np.zeros(10))
    tb1 = types.Table(col_dict={"dset1`": types.Array(np.ones(10))})
    store = lh5.LH5Store()
    store.write(array1, "my_table", f"{tmptestdir}/write_object_append_column_test.lh5")
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        store.write(
            tb1,
            "my_table",
            f"{tmptestdir}/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column to an array

    # Try to append a table that has a same key as the old table
    if Path(f"{tmptestdir}/write_object_append_column_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_append_column_test.lh5").unlink()

    tb1 = types.Table(
        col_dict={
            "dset1": types.Array(np.zeros(10)),
            "dset2": types.Array(np.zeros(10)),
        }
    )
    tb2 = types.Table(
        col_dict={"dset2": types.Array(np.ones(10))}
    )  # Same field name, different values
    store = lh5.LH5Store()
    store.write(tb1, "my_table", f"{tmptestdir}/write_object_append_column_test.lh5")
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        store.write(
            tb2,
            "my_table",
            f"{tmptestdir}/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column with a same field

    # try appending a column that is larger than one that exists
    if Path(f"{tmptestdir}/write_object_append_column_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_append_column_test.lh5").unlink()

    tb1 = types.Table(col_dict={"dset1": types.Array(np.zeros(10))})
    tb2 = types.Table(
        col_dict={"dset2": types.Array(np.ones(20))}
    )  # different field name, different size
    store = lh5.LH5Store()
    store.write(tb1, "my_table", f"{tmptestdir}/write_object_append_column_test.lh5")
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        store.write(
            tb2,
            "my_table",
            f"{tmptestdir}/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column with a different field size

    # Finally successfully append a column
    if Path(f"{tmptestdir}/write_object_append_column_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_append_column_test.lh5").unlink()

    tb1 = types.Table(col_dict={"dset1": types.Array(np.zeros(10))})
    tb2 = types.Table(
        col_dict={"dset2": types.Array(np.ones(10))}
    )  # different field name, different size
    store = lh5.LH5Store()
    store.write(
        tb1,
        "my_table",
        f"{tmptestdir}/write_object_append_column_test.lh5",
        group="my_group",
    )
    store.write(
        tb2,
        "my_table",
        f"{tmptestdir}/write_object_append_column_test.lh5",
        group="my_group",
        wo_mode="append_column",
    )

    # Now, check that the data were appended
    tb_dat = store.read(
        "my_group/my_table", f"{tmptestdir}/write_object_append_column_test.lh5"
    )
    assert isinstance(tb_dat, types.Table)
    assert np.array_equal(tb_dat["dset1"].nda, np.zeros(10))
    assert np.array_equal(tb_dat["dset2"].nda, np.ones(10))


# Test writing and reading histograms.
def test_write_histogram(caplog, tmptestdir):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    # Start with an types.Histogram
    if Path(f"{tmptestdir}/write_histogram_test.lh5").exists():
        Path(f"{tmptestdir}/write_histogram_test.lh5").unlink()

    h1 = types.Histogram(
        np.array([[1, 1], [1, 1]]), (np.array([0, 1, 2]), np.array([2.1, 2.2, 2.3]))
    )
    h2 = types.Histogram(
        np.array([[10, 10], [10, 10]]),
        (np.array([2, 3, 4]), np.array([5, 6, 7])),
        isdensity=True,
    )
    h2.binning[0]["binedges"].attrs["units"] = "ns"

    # Same field name, different values
    store = lh5.LH5Store()
    # "appending" to a non-existing histogram should work.
    store.write(
        h1,
        "my_histogram",
        f"{tmptestdir}/write_histogram_test.lh5",
        group="my_group",
        wo_mode="append",
    )
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        # appending to an existing histogram should not work.
        store.write(
            h1,
            "my_histogram",
            f"{tmptestdir}/write_histogram_test.lh5",
            group="my_group",
            wo_mode="append",
        )
    store.write(
        h2,
        "my_histogram",
        f"{tmptestdir}/write_histogram_test.lh5",
        wo_mode="overwrite",
        group="my_group",
    )

    # Now, check that the data were overwritten
    h3 = store.read("my_group/my_histogram", f"{tmptestdir}/write_histogram_test.lh5")
    assert np.array_equal(h3.weights.nda, np.array([[10, 10], [10, 10]]))
    assert h3.binning[0].edges[0] == 2
    assert h3.binning[1].edges[-1] == 7
    assert h3.isdensity
    assert h3.binning[0].get_binedgeattrs() == {"units": "ns"}

    # Now, check that writing with other modes throws.
    for disallowed_wo_mode in ["append", "append_column"]:
        with pytest.raises(lh5.exceptions.LH5EncodeError):
            store.write(
                h2,
                "my_histogram",
                f"{tmptestdir}/write_histogram_test.lh5",
                wo_mode=disallowed_wo_mode,
                group="my_group",
            )
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        store.write(
            h2,
            "my_histogram",
            f"{tmptestdir}/write_histogram_test.lh5",
            wo_mode="overwrite",
            write_start=1,
            group="my_group",
        )


def test_write_histogram_variable(caplog, tmptestdir):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    # Start with an types.Histogram
    if Path(f"{tmptestdir}/write_histogram_test.lh5").exists():
        Path(f"{tmptestdir}/write_histogram_test.lh5").unlink()

    h1 = types.Histogram(
        np.array([[1, 1], [1, 1]]), (np.array([0, 1.2, 2]), np.array([2.1, 2.5, 2.3]))
    )
    h2 = types.Histogram(
        np.array([[10, 10], [10, 10]]),
        (np.array([2, 3.5, 4]), np.array([5, 6.5, 7])),
        isdensity=True,
        attrs={"testattr": "test"},
    )
    h2["weights"].attrs["weightattr"] = "testweight"
    h2.binning[0].attrs["binningattr"] = "testbinning"

    # Same field name, different values
    store = lh5.LH5Store()
    store.write(
        h1,
        "my_histogram",
        f"{tmptestdir}/write_histogram_test.lh5",
        group="my_group",
        wo_mode="write_safe",
    )
    store.write(
        h2,
        "my_histogram",
        f"{tmptestdir}/write_histogram_test.lh5",
        wo_mode="overwrite",
        group="my_group",
    )

    # Now, check that the data were overwritten
    h3 = store.read("my_group/my_histogram", f"{tmptestdir}/write_histogram_test.lh5")
    assert np.array_equal(h3.weights.nda, np.array([[10, 10], [10, 10]]))
    assert np.array_equal(h3.binning[0].edges, np.array([2, 3.5, 4]))
    with pytest.raises(TypeError):
        x = h3.binning[0].first
    with pytest.raises(TypeError):
        x = h3.binning[1].last  # noqa: F841
    assert not h3.binning[0].is_range
    assert h3.isdensity

    # ensure that reading back attrs not only on binedges works.
    assert h3.attrs["testattr"] == "test"
    assert h3["weights"].attrs["weightattr"] == "testweight"
    assert h3.binning[0].attrs["binningattr"] == "testbinning"



from __future__ import annotations

from .read.array import (
    _h5_read_array,
    _h5_read_array_of_equalsized_arrays,
    _h5_read_fixedsize_array,
    _h5_read_ndarray,
)
from .read.composite import (
    _h5_read_histogram,
    _h5_read_lgdo,
    _h5_read_struct,
    _h5_read_table,
)
from .read.encoded import (
    _h5_read_array_of_encoded_equalsized_arrays,
    _h5_read_vector_of_encoded_vectors,
)
from .read.scalar import _h5_read_scalar
from .read.vector_of_vectors import _h5_read_vector_of_vectors
from .write.array import _h5_write_array
from .write.composite import _h5_write_lgdo, _h5_write_struct
from .write.scalar import _h5_write_scalar
from .write.vector_of_vectors import _h5_write_vector_of_vectors

__all__ = [
    "_h5_read_array",
    "_h5_read_array_of_encoded_equalsized_arrays",
    "_h5_read_array_of_equalsized_arrays",
    "_h5_read_encoded_array",
    "_h5_read_fixedsize_array",
    "_h5_read_histogram",
    "_h5_read_lgdo",
    "_h5_read_ndarray",
    "_h5_read_scalar",
    "_h5_read_struct",
    "_h5_read_table",
    "_h5_read_vector_of_encoded_vectors",
    "_h5_read_vector_of_vectors",
    "_h5_write_array",
    "_h5_write_lgdo",
    "_h5_write_scalar",
    "_h5_write_struct",
    "_h5_write_vector_of_vectors",
]



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Collection, Mapping

import h5py
import numpy as np

from .... import types
from ... import datatype
from ...exceptions import LH5DecodeError
from . import scalar

log = logging.getLogger(__name__)


def check_obj_buf_attrs(attrs, new_attrs, fname, oname):
    if set(attrs.keys()) != set(new_attrs.keys()):
        msg = (
            f"existing buffer and new data chunk have different attributes: "
            f"obj_buf.attrs={attrs} != {fname}[{oname}].attrs={new_attrs}"
        )
        raise LH5DecodeError(msg, fname, oname)


def build_field_mask(field_mask: Mapping[str, bool] | Collection[str]) -> defaultdict:
    # check field_mask and make it a default dict
    if field_mask is None:
        return defaultdict(lambda: True)
    if isinstance(field_mask, dict):
        default = True
        if len(field_mask) > 0:
            default = not field_mask[next(iter(field_mask.keys()))]
        return defaultdict(lambda: default, field_mask)
    if isinstance(field_mask, (list, tuple, set)):
        return defaultdict(bool, dict.fromkeys(field_mask, True))
    if isinstance(field_mask, defaultdict):
        return field_mask
    msg = "bad field_mask type"
    raise ValueError(msg, type(field_mask).__name__)


def eval_field_mask(
    field_mask: defaultdict, all_fields: list[str]
) -> list[tuple(str, defaultdict)]:
    """Get list of fields that need to be loaded along with a sub-field-mask
    in case we have a nested Table"""

    if field_mask is None:
        return all_fields

    this_field_mask = defaultdict(field_mask.default_factory)
    sub_field_masks = {}

    for key, val in field_mask.items():
        field = key.strip("/")
        pos = field.find("/")
        if pos < 0:
            this_field_mask[field] = val
        else:
            sub_field = field[pos + 1 :]
            field = field[:pos]
            this_field_mask[field] = True
            sub_mask = sub_field_masks.setdefault(
                field, defaultdict(field_mask.default_factory)
            )
            sub_mask[sub_field] = val

    return [
        (field, sub_field_masks.get(field))
        for field in all_fields
        if this_field_mask[field]
    ]


def read_attrs(h5o, fname, oname):
    """Read all attributes for an hdf5 dataset or group using low level API
    and return them as a dict. Assume all are strings or scalar types."""
    attrs = {}
    for i_attr in range(h5py.h5a.get_num_attrs(h5o)):
        h5a = h5py.h5a.open(h5o, index=i_attr)
        name = h5a.get_name().decode()
        if h5a.shape != ():
            msg = f"attribute {oname} is not a string or scalar"
            raise LH5DecodeError(msg, fname, oname)
        val = np.empty((), h5a.dtype)
        h5a.read(val)
        if h5a.get_type().get_class() == h5py.h5t.STRING:
            attrs[name] = val.item().decode()
        else:
            attrs[name] = val.item()
        h5a.close()
    return attrs


def read_n_rows(h5o, fname, oname):
    """Read number of rows in LH5 object"""
    if not h5py.h5a.exists(h5o, b"datatype"):
        msg = "missing 'datatype' attribute"
        raise LH5DecodeError(msg, fname, oname)

    h5a = h5py.h5a.open(h5o, b"datatype")
    type_attr = np.empty((), h5a.dtype)
    h5a.read(type_attr)
    type_attr = type_attr.item().decode()
    lgdotype = datatype.datatype(type_attr)

    # scalars are dim-0 datasets
    if lgdotype is types.Scalar:
        return None

    # structs don't have rows
    if lgdotype is types.Struct:
        return None

    # tables should have elements with all the same length
    if lgdotype is types.Table:
        # read out each of the fields
        rows_read = None
        for field in datatype.get_struct_fields(type_attr):
            obj = h5py.h5o.open(h5o, field.encode())
            n_rows_read = read_n_rows(obj, fname, field)
            obj.close()
            if not rows_read:
                rows_read = n_rows_read
            elif rows_read != n_rows_read:
                log.warning(
                    f"'{field}' field in table '{oname}' has {rows_read} rows, "
                    f"{n_rows_read} was expected"
                )

        return rows_read

    # length of vector of vectors is the length of its cumulative_length
    if lgdotype is types.VectorOfVectors:
        obj = h5py.h5o.open(h5o, b"cumulative_length")
        n_rows = read_n_rows(obj, fname, "cumulative_length")
        obj.close()
        return n_rows

    # length of vector of encoded vectors is the length of its decoded_size
    if lgdotype in (types.VectorOfEncodedVectors, types.ArrayOfEncodedEqualSizedArrays):
        obj = h5py.h5o.open(h5o, b"encoded_data")
        n_rows = read_n_rows(obj, fname, "encoded_data")
        obj.close()
        return n_rows

    # return array length (without reading the array!)
    if issubclass(lgdotype, types.Array):
        # compute the number of rows to read
        return h5o.get_space().shape[0]

    msg = f"don't know how to read rows of LGDO {lgdotype.__name__}"
    raise LH5DecodeError(msg, fname, oname)


def read_size_in_bytes(h5o, fname, oname, field_mask=None):
    """Read number size in LH5 object in memory (in B)"""
    if not h5py.h5a.exists(h5o, b"datatype"):
        msg = "missing 'datatype' attribute"
        raise LH5DecodeError(msg, fname, oname)

    h5a = h5py.h5a.open(h5o, b"datatype")
    type_attr = np.empty((), h5a.dtype)
    h5a.read(type_attr)
    type_attr = type_attr.item().decode()
    lgdotype = datatype.datatype(type_attr)
    field_mask = build_field_mask(field_mask)

    # scalars are dim-0 datasets
    if lgdotype in (
        types.Scalar,
        types.Array,
        types.ArrayOfEqualSizedArrays,
        types.FixedSizeArray,
    ):
        return int(np.prod(h5o.shape) * h5o.dtype.itemsize)

    # tables should have elements with all the same length
    if lgdotype in (
        types.Struct,
        types.Histogram,
        types.Histogram.Axis,
        types.Table,
        types.WaveformTable,
    ):
        # read out each of the fields
        size = 0
        all_fields = datatype.get_struct_fields(type_attr)
        selected_fields = eval_field_mask(field_mask, all_fields)
        for field, submask in selected_fields:
            obj = h5py.h5o.open(h5o, field.encode())
            size += read_size_in_bytes(obj, fname, field, submask)
            obj.close()
        return size

    # length of vector of vectors is the length of its cumulative_length
    if lgdotype is types.VectorOfVectors:
        size = 0
        obj = h5py.h5o.open(h5o, b"cumulative_length")
        size += read_size_in_bytes(obj, fname, "cumulative_length")
        obj.close()
        obj = h5py.h5o.open(h5o, b"flattened_data")
        size += read_size_in_bytes(obj, fname, "flattened_data")
        obj.close()
        return size

    # length of vector of encoded vectors is the length of its decoded_size
    if lgdotype is types.ArrayOfEncodedEqualSizedArrays:
        obj = h5py.h5o.open(h5o, b"decoded_size")
        size = scalar._h5_read_scalar(obj, fname, "decoded_size")[0].value
        obj.close()

        obj = h5py.h5o.open(h5o, b"encoded_data")
        cl = h5py.h5o.open(obj, b"cumulative_length")
        size *= cl.shape[0]
        size *= 4  # TODO: UPDATE WHEN CODECS SUPPORT MORE DTYPES
        obj.close()

        return size

    msg = f"don't know how to read size of LGDO {lgdotype.__name__}"
    raise LH5DecodeError(msg, fname, oname)



from __future__ import annotations

from lgdo import cli, lh5


def test_lh5ls(lgnd_test_data):
    cli.lh5ls(
        [
            "-a",
            lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
            "geds/raw",
        ]
    )


def test_lh5concat(lgnd_test_data, tmptestdir):
    infile1 = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_raw.lh5"
    )
    infile2 = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/phy/p03/r001/l200-p03-r001-phy-20230322T170202Z-tier_raw.lh5"
    )
    outfile = f"{tmptestdir}/out.lh5"

    cli.lh5concat_cli(
        ["--output", outfile, "--debug", "--overwrite", "--", infile1, infile2]
    )

    assert lh5.ls(outfile) == [
        "ch1057600",
        "ch1059201",
        "ch1062405",
        "ch1084803",
        "ch1084804",
        "ch1121600",
    ]



from __future__ import annotations

import bisect
import inspect
import sys
from collections.abc import Mapping, Sequence
from contextlib import suppress
from typing import Any

import h5py
import numpy as np
from numpy.typing import ArrayLike

from .. import types
from . import _serializers
from .utils import read_n_rows


def read(
    name: str,
    lh5_file: str | h5py.File | Sequence[str | h5py.File],
    start_row: int = 0,
    n_rows: int = sys.maxsize,
    idx: ArrayLike = None,
    use_h5idx: bool = False,
    field_mask: Mapping[str, bool] | Sequence[str] | None = None,
    obj_buf: types.LGDO = None,
    obj_buf_start: int = 0,
    decompress: bool = True,
    locking: bool = False,
) -> types.LGDO | tuple[types.LGDO, int]:
    """Read LH5 object data from a file.

    Note
    ----
    Use the ``idx`` parameter to read out particular rows of the data. The
    ``use_h5idx`` flag controls whether *only* those rows are read from
    disk or if the rows are indexed after reading the entire object.
    Reading individual rows can be orders of magnitude slower than reading
    the whole object and then indexing the desired rows. The default
    behavior (``use_h5idx=False``) is to use slightly more memory for a
    much faster read. See `legend-pydataobj/issues/#29
    <https://github.com/legend-exp/legend-pydataobj/issues/29>`_ for
    additional information.

    Parameters
    ----------
    name
        Name of the LH5 object to be read (including its group path).
    lh5_file
        The file(s) containing the object to be read out. If a list of
        files, array-like object data will be concatenated into the output
        object.
    start_row
        Starting entry for the object read (for array-like objects). For a
        list of files, only applies to the first file.
    n_rows
        The maximum number of rows to read (for array-like objects). The
        actual number of rows read will be returned as one of the return
        values (see below).
    idx
        For NumPy-style "fancying indexing" for the read to select only
        some rows, e.g. after applying some cuts to particular columns.
        Only selection along the first axis is supported, so tuple
        arguments must be one-tuples.  If `n_rows` is not false, `idx` will
        be truncated to `n_rows` before reading. To use with a list of
        files, can pass in a list of `idx`'s (one for each file) or use a
        long contiguous list (e.g. built from a previous identical read).
        If used in conjunction with `start_row` and `n_rows`, will be
        sliced to obey those constraints, where `n_rows` is interpreted as
        the (max) number of *selected* values (in `idx`) to be read out.
        Note that the ``use_h5idx`` parameter controls some behaviour of
        the read and that the default behavior (``use_h5idx=False``)
        prioritizes speed over a small memory penalty.
    use_h5idx
        ``True`` will directly pass the ``idx`` parameter to the underlying
        :mod:`h5py` call such that only the selected rows are read directly
        into memory, which conserves memory at the cost of speed. There can
        be a significant penalty to speed for larger files (1 - 2 orders of
        magnitude longer time).  ``False`` (default) will read the entire
        object into memory before performing the indexing. The default is
        much faster but requires additional memory, though a relatively
        small amount in the typical use case. It is recommended to leave
        this parameter as its default.
    field_mask
        For tables and structs, determines which fields get read out.
        Only applies to immediate fields of the requested objects. If a dict
        is used, a default dict will be made with the default set to the
        opposite of the first element in the dict. This way if one specifies
        a few fields at ``False``, all but those fields will be read out,
        while if one specifies just a few fields as ``True``, only those
        fields will be read out. If a list is provided, the listed fields
        will be set to ``True``, while the rest will default to ``False``.
    obj_buf
        Read directly into memory provided in `obj_buf`. Note: the buffer
        will be resized to accommodate the data retrieved.
    obj_buf_start
        Start location in ``obj_buf`` for read. For concatenating data to
        array-like objects.
    decompress
        Decompress data encoded with LGDO's compression routines right
        after reading. The option has no effect on data encoded with HDF5
        built-in filters, which is always decompressed upstream by HDF5.
    locking
        Lock HDF5 file while reading

    Returns
    -------
    object
        the read-out object
    """
    if isinstance(lh5_file, h5py.File):
        lh5_obj = lh5_file[name]
    elif isinstance(lh5_file, str):
        lh5_file = h5py.File(lh5_file, mode="r", locking=locking)
        try:
            lh5_obj = lh5_file[name]
        except KeyError as ke:
            err = f"Object {name} not found in file {lh5_file.filename}"
            raise KeyError(err) from ke
    else:
        if obj_buf is not None:
            obj_buf.resize(obj_buf_start)
        else:
            obj_buf_start = 0

        for i, h5f in enumerate(lh5_file):
            if (
                isinstance(idx, (list, tuple))
                and len(idx) > 0
                and not np.isscalar(idx[0])
            ):
                # a list of lists: must be one per file
                idx_i = idx[i]
            elif idx is not None:
                # make idx a proper tuple if it's not one already
                if not (isinstance(idx, tuple) and len(idx) == 1):
                    idx = (idx,)
                # idx is a long continuous array
                n_rows_i = read_n_rows(name, h5f)
                # find the length of the subset of idx that contains indices
                # that are less than n_rows_i
                n_rows_to_read_i = bisect.bisect_left(idx[0], n_rows_i)
                # now split idx into idx_i and the remainder
                idx_i = np.array(idx[0])[:n_rows_to_read_i]
                idx = np.array(idx[0])[n_rows_to_read_i:] - n_rows_i
            else:
                idx_i = None

            obj_buf_start_i = len(obj_buf) if obj_buf else 0
            n_rows_i = n_rows - (obj_buf_start_i - obj_buf_start)

            obj_buf = read(
                name,
                h5f,
                start_row if i == 0 else 0,
                n_rows_i,
                idx_i,
                use_h5idx,
                field_mask,
                obj_buf,
                obj_buf_start_i,
                decompress,
            )

            if obj_buf is None or (len(obj_buf) - obj_buf_start) >= n_rows:
                return obj_buf
        return obj_buf

    if isinstance(idx, (list, tuple)) and len(idx) > 0 and not np.isscalar(idx[0]):
        idx = idx[0]
    if isinstance(idx, np.ndarray) and idx.dtype == np.dtype("?"):
        idx = np.where(idx)[0]

    obj, n_rows_read = _serializers._h5_read_lgdo(
        lh5_obj.id,
        lh5_obj.file.filename,
        lh5_obj.name,
        start_row=start_row,
        n_rows=n_rows,
        idx=idx,
        use_h5idx=use_h5idx,
        field_mask=field_mask,
        obj_buf=obj_buf,
        obj_buf_start=obj_buf_start,
        decompress=decompress,
    )
    with suppress(AttributeError):
        obj.resize(obj_buf_start + n_rows_read)

    return obj


def write(
    obj: types.LGDO,
    name: str,
    lh5_file: str | h5py.File,
    group: str | h5py.Group = "/",
    start_row: int = 0,
    n_rows: int | None = None,
    wo_mode: str = "append",
    write_start: int = 0,
    page_buffer: int = 0,
    **h5py_kwargs,
) -> None:
    """Write an LGDO into an LH5 file.

    If the `obj` :class:`.LGDO` has a `compression` attribute, its value is
    interpreted as the algorithm to be used to compress `obj` before
    writing to disk. The type of `compression` can be:

    string, kwargs dictionary, hdf5plugin filter
      interpreted as the name of a built-in or custom `HDF5 compression
      filter <https://docs.h5py.org/en/stable/high/dataset.html#filter-pipeline>`_
      (``"gzip"``, ``"lzf"``, :mod:`hdf5plugin` filter object etc.) and
      passed directly to :meth:`h5py.Group.create_dataset`.

    :class:`.WaveformCodec` object
      If `obj` is a :class:`.WaveformTable` and ``obj.values`` holds the
      attribute, compress ``values`` using this algorithm. More
      documentation about the supported waveform compression algorithms at
      :mod:`.lgdo.compression`.

    If the `obj` :class:`.LGDO` has a `hdf5_settings` attribute holding a
    dictionary, it is interpreted as a list of keyword arguments to be
    forwarded directly to :meth:`h5py.Group.create_dataset` (exactly like
    the first format of `compression` above). This is the preferred way to
    specify HDF5 dataset options such as chunking etc. If compression
    options are specified, they take precedence over those set with the
    `compression` attribute.

    Note
    ----------
    The `compression` LGDO attribute takes precedence over the default HDF5
    compression settings. The `hdf5_settings` attribute takes precedence
    over `compression`. These attributes are not written to disk.

    Note
    ----------
    HDF5 compression is skipped for the `encoded_data.flattened_data`
    dataset of :class:`.VectorOfEncodedVectors` and
    :class:`.ArrayOfEncodedEqualSizedArrays`.

    Parameters
    ----------
    obj
        LH5 object. if object is array-like, writes `n_rows` starting from
        `start_row` in `obj`.
    name
        name of the object in the output HDF5 file.
    lh5_file
        HDF5 file name or :class:`h5py.File` object.
    group
        HDF5 group name or :class:`h5py.Group` object in which `obj` should
        be written.
    start_row
        first row in `obj` to be written.
    n_rows
        number of rows in `obj` to be written.
    wo_mode
        - ``write_safe`` or ``w``: only proceed with writing if the
          object does not already exist in the file.
        - ``append`` or ``a``: append along axis 0 (the first dimension)
          of array-like objects and array-like subfields of structs.
          :class:`~.lgdo.scalar.Scalar` objects get overwritten.
        - ``overwrite`` or ``o``: replace data in the file if present,
          starting from `write_start`. Note: overwriting with `write_start` =
          end of array is the same as ``append``.
        - ``overwrite_file`` or ``of``: delete file if present prior to
          writing to it. `write_start` should be 0 (its ignored).
        - ``append_column`` or ``ac``: append columns from an
          :class:`~.lgdo.table.Table` `obj` only if there is an existing
          :class:`~.lgdo.table.Table` in the `lh5_file` with the same
          `name` and :class:`~.lgdo.table.Table.size`. If the sizes don't
          match, or if there are matching fields, it errors out.
    write_start
        row in the output file (if already existing) to start overwriting
        from.
    page_buffer
        enable paged aggregation with a buffer of this size in bytes
        Only used when creating a new file. Useful when writing a file
        with a large number of small datasets. This is a short-hand for
        ``(fs_stragety="page", fs_pagesize=[page_buffer])``
    **h5py_kwargs
        additional keyword arguments forwarded to
        :meth:`h5py.Group.create_dataset` to specify, for example, an HDF5
        compression filter to be applied before writing non-scalar
        datasets. **Note: `compression` Ignored if compression is specified
        as an `obj` attribute.**
    """
    if wo_mode in ("w", "write", "of", "overwrite_file"):
        h5py_kwargs.update(
            {
                "fs_strategy": "page",
                "fs_page_size": page_buffer,
            }
        )
    return _serializers._h5_write_lgdo(
        obj,
        name,
        lh5_file,
        group=group,
        start_row=start_row,
        n_rows=n_rows,
        wo_mode=wo_mode,
        write_start=write_start,
        **h5py_kwargs,
    )


def read_as(
    name: str,
    lh5_file: str | h5py.File | Sequence[str | h5py.File],
    library: str,
    **kwargs,
) -> Any:
    """Read LH5 data from disk straight into a third-party data format view.

    This function is nothing more than a shortcut chained call to
    :func:`.read` and to :meth:`.LGDO.view_as`.

    Parameters
    ----------
    name
        LH5 object name on disk.
    lh5_file
        LH5 file name.
    library
        string ID of the third-party data format library (``np``, ``pd``,
        ``ak``, etc).

    See Also
    --------
    .read, .LGDO.view_as
    """
    # determine which keyword arguments should be forwarded to read() and which
    # should be forwarded to view_as()
    read_kwargs = inspect.signature(read).parameters.keys()

    kwargs1 = {}
    kwargs2 = {}
    for k, v in kwargs.items():
        if k in read_kwargs:
            kwargs1[k] = v
        else:
            kwargs2[k] = v

    # read the LGDO from disk
    # NOTE: providing a buffer does not make much sense
    obj = read(name, lh5_file, **kwargs1)

    if isinstance(obj, tuple):
        obj = obj[0]

    # and finally return a view
    return obj.view_as(library, **kwargs2)



"""Routines from reading and writing LEGEND Data Objects in HDF5 files.
Currently the primary on-disk format for LGDO object is LEGEND HDF5 (LH5) files. IO
is done via the class :class:`.store.LH5Store`. LH5 files can also be
browsed easily in python like any `HDF5 <https://www.hdfgroup.org>`_ file using
`h5py <https://www.h5py.org>`_.
"""

from __future__ import annotations

from ._serializers.write.array import DEFAULT_HDF5_SETTINGS
from .core import read, read_as, write
from .iterator import LH5Iterator
from .store import LH5Store
from .tools import ls, show
from .utils import read_n_rows

__all__ = [
    "DEFAULT_HDF5_SETTINGS",
    "LH5Iterator",
    "LH5Store",
    "concat",
    "ls",
    "read",
    "read_as",
    "read_n_rows",
    "show",
    "write",
]



from __future__ import annotations

import numpy as np

import lgdo


def test_datatype_name():
    array = lgdo.FixedSizeArray()
    assert array.datatype_name() == "fixedsize_array"


def test_form_datatype():
    array = lgdo.FixedSizeArray()
    assert array.form_datatype() == "fixedsize_array<0>{real}"


def test_init():
    attrs = {"attr1": 1}
    array = lgdo.FixedSizeArray(shape=(3,), dtype=np.float32, fill_val=42, attrs=attrs)
    assert (array.nda == np.full((3,), 42, np.float32)).all()
    assert array.attrs == attrs | {"datatype": "fixedsize_array<1>{real}"}



from __future__ import annotations

from lgdo import ArrayOfEncodedEqualSizedArrays, compression
from lgdo.compression import RadwareSigcompress
from lgdo.compression.generic import _is_codec


def test_encode_decode_array(wftable):
    result = compression.encode(
        wftable.values, codec=RadwareSigcompress(codec_shift=-32768)
    )
    assert isinstance(result, ArrayOfEncodedEqualSizedArrays)
    assert len(result) == len(wftable)
    assert result.attrs["codec"] == "radware_sigcompress"
    assert result.attrs["codec_shift"] == -32768

    compression.decode(result)


def test_is_codec():
    assert _is_codec("radware_sigcompress", RadwareSigcompress)
